from __future__ import print_function
import scipy as sp
import numpy as np
from .TomoFunctionsHelpers import *
import numpy.random as rand
from scipy.linalg import cholesky

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""


# # # # # # # # # # # # #
# TOMOGRAPHY CALCULATE  #
# # # # # # # # # # # # #



"""
log_likelyhood(t, coincidences, accidentals, m, prediction)
Desc: This is the log likelyhood function. It used in bayesian tomography.

Parameters
----------
t : ndarray
    T values of the current predicted state.
coincidences : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The counts of the tomography.
accidentals : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The singles values of the tomography. Used for accidental correction.
m : ndarray with shape = (2^numQubits, 2^numQubits, number of measurements)
    The measurements of the tomography in density matrix form.
prediction : ndarray
    Predicted counts from the predicted state.
overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit. (optional)
    The relative weights of each measurment. Used for drift correction.
    
Returns
-------
val : float
    value of the optimization function.
"""
def log_likelyhood(intensity, givenState, coincidences, measurements, accidentals, overall_norms=-1):
    # If overall_norms not given then assume uniform
    if not isinstance(overall_norms, np.ndarray):
        overall_norms = np.ones(coincidences.shape[0])
    elif not (len(overall_norms.shape) == 1 and overall_norms.shape[0] == coincidences.shape[0]):
        raise ValueError("Invalid intensities array")
    # Convert to densities if given tvals
    if (len(givenState.shape) == 1):
        givenState = t_to_density(givenState)
    # Calculate expected averages
    Averages = np.zeros_like(coincidences, dtype=np.float)
    givenState = intensity*givenState
    for j in range(len(Averages)):
        Averages[j] = overall_norms[j] * np.real(np.trace(np.matmul(measurements[j], givenState))) + \
                      accidentals[j]
        # Avoid dividing by zero for pure states
        if (Averages[j] == 0):
            Averages[j] == 1
    val = (Averages - coincidences) ** 2 / (2 * Averages)
    return sum(val)


"""
density2tm(rhog)
Desc: Converts a density matrix into a lower t matrix.

Parameters
----------
rhog : ndarray
    Array to be converted.

Returns
-------
tm : ndarray
    Lower t matrix defining the input matrix.
"""
def density2tm(rhog):
    return np.linalg.cholesky(rhog)

"""
density2t(rhog)
Desc: Converts a density matrix into a list of t values

Parameters
----------
rhog : ndarray
    Array to be converted.

Returns
-------
t : ndarray
    List of t values defining the input matrix.
"""
def density2t(rhog):
    tm = density2tm(rhog)
    d = len(tm)
    idx = 0
    cur_length = d
    t = np.zeros(d**2)
    for j in range(d):
        t[np.arange(idx, idx+cur_length)] = np.real(np.diag(tm, -j))
        idx = idx + cur_length
        if j > 0:
            t[np.arange(idx, idx+cur_length)] = np.imag(np.diag(tm, -j))
            idx = idx + cur_length
        cur_length -= 1
    return t

"""
toDensity(psiMat)
Desc: Converts a pure state into a density matrix.

Parameters
----------
psiMat : ndarray
    Pure state to be converted.

Returns
-------
rhog: ndarray
    Density Matrix of the input state.
    """
def toDensity(psiMat):
    if not isStateVector(psiMat):
        raise ValueError("Invalid input state with shape " + str(psiMat.shape))
    psiMat = np.outer(psiMat, psiMat.conj())
    return psiMat / np.trace(psiMat)


"""
t_matrix(t)
Desc: Converts a list of t values to an lower t matrix.

Parameters
----------
t : ndarray
    List of t values converted.

Returns
-------
tm : ndarray
    Lower t matrix.
    """
def t_matrix(t):
    d = int(np.sqrt(len(t)))
    idx = 0
    cur_length = d
    tm = np.zeros([d, d])
    for j in range(int(d)):
        tm = tm + 1*np.diag(t[np.arange(idx, idx+cur_length)], -j)
        idx = idx + cur_length
        if j > 0:
            tm = tm + 1j*np.diag(t[np.arange(idx, idx+cur_length)], -j)
            idx = idx + cur_length
        cur_length -= 1
    return tm

"""
t_to_density(t)
Desc: Converts a list of t values to a density matrix.

Parameters
----------
t : ndarray
    List of t values converted.
normalize : bool
    Set to true if you want the density matrix to be normalized by its trace. 
    Default is True.

Returns
-------
rhog : ndarray
    Density Matrix.
    """
def t_to_density(t,normalize=True):
    tm = t_matrix(t)
    rhog = np.dot(tm,tm.T.conj())
    if normalize:
        rhog = rhog / np.trace(rhog)
    return rhog

"""
fidelity(state1, state2)
Desc: Calculates the fidelity between the two input states.

Parameters
----------
state1, state2 : ndarray
    Input arrays to calculate the fidelity between. Can be pure states or density matrices.

Returns
-------
val : float
    The calculated fidelity.
"""
def fidelity(state1, state2):
    rho1 = state1
    rho2 = state2
    pure = 0
    if isStateVector(state1):
        rho1 = toDensity(state1)
        pure = 1
    elif state1.shape[1] == state1.shape[0]:
        rho1 = state1
    else:
        raise ValueError("State1 is not a vector or density matrix")

    if isStateVector(state2):
        rho2 = toDensity(state2)
        pure = 1
    elif state2.shape[1] == state2.shape[0]:
        rho2 = state2
    else:
        raise ValueError("State2 is not a vector or density matrix")

    rho1 = rho1 /np.trace(rho1)
    rho2 = rho2 / np.trace(rho2)

    if pure:
        val = np.trace(np.dot(rho1, rho2))
    else:
        tmp = sp.linalg.sqrtm(rho1)
        a = np.dot(tmp, np.dot(rho2, tmp))
        val = (np.trace(sp.linalg.sqrtm(a)))**2
    val = np.real(val)

    # when comparing 2 identical pure state, it will get a value larger than 1,
    if val > 1:
        if val - 1 < 0.00001:
            val = 1.0
        else:
            print("Fidelity larger than 1.")

    return val


# # # # # # # #
# PROPERTIES  #
# # # # # # # #


"""
concurrence(rhog)
Desc: Calculates the concurrence of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state.

Returns
-------
val : float
    The calculated concurrence.
   
Other Properties
 -------------- 
entropy;linear_entropy;negativity;purity;tangle

See Also
 ------ 
err_functions;getProperties
"""
def concurrence(rhog):
    if rhog.shape[0] == 4:
        if isStateVector(rhog):
            rhog = toDensity(rhog)
        if min(rhog.shape) == 1:
            rhog = np.dot(rhog.conj(), rhog.transpose())

        zz = np.array([[0, 0, 0, -1], [0, 0, 1, 0], [0, 1, 0, 0], [-1, 0, 0, 0]])
        rr = np.dot(rhog, np.dot(zz, np.dot(rhog.conj(), zz)))
        r = np.linalg.eig(rr)[0]
        # left = np.linalg.inv(right)
        r = np.real(r)

        tmp = np.sort(np.sqrt(r+0j))
        val = np.real(tmp[3]-tmp[2]-tmp[1]-tmp[0])
        val = np.max([val, 0])

        return val
    else:
        return 'NA'


"""
tangle(rhog)
Desc: Calculates the tangle of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state. Tangle is calculated by squaring the concurrence.

Returns
-------
val : float
    The calculated tangle.

Other Properties
 -------------- 
entropy;linear_entropy;negativity;purity;concurrence

See Also
 ------ 
err_functions;getProperties
"""
def tangle(rhog):
    if rhog.shape[0] == 4:
        if isStateVector(rhog):
            rhog = toDensity(rhog)
        c = concurrence(rhog)
        val = c ** 2
        return val
    else:
        return 'NA'

"""
entropy(rhog)
Desc: Calculates the Von Neumann Entropy of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state.

Returns
-------
val : float
    The calculated entropy.

Other Properties
 -------------- 
concurrence;linear_entropy;negativity;purity;tangle

See Also
 ------ 
err_functions;getProperties
"""
def entropy(rhog):
    if isStateVector(rhog):
        rhog = toDensity(rhog)
    d = np.linalg.eig(rhog)[0]
    e = np.real(d)
    val = 0
    for a in range(len(e)):
        if e[a] > 0:
            val = val-e[a]*np.log2(e[a])
    return val

"""
linear_entropy(rhog)
Desc: Calculates the linear entropy of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state.

Returns
-------
val : float
    The calculated linear entropy ranging from 0 to 1/(2^numQubits).
    A value of zero corresponds to a completly pure state.

Other Properties
 -------------- 
entropy;concurrence;negativity;purity;tangle

See Also
 ------ 
err_functions;getProperties
"""
def linear_entropy(rhog):
    if isStateVector(rhog):
        rhog = toDensity(rhog)
    return 1- purity(rhog)


"""
negativity(rhog)
Desc: Calculates the negativity of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state.

Returns
-------
val : float
    The calculated negativity.

Other Properties
 -------------- 
entropy;linear_entropy;concurrence;purity;tangle

See Also
 ------ 
err_functions;getProperties
"""
def negativity(rhog):
    if isStateVector(rhog):
        rhog = toDensity(rhog)
    if rhog.shape[0] == 4:
        rho1 = partial_transpose(rhog)
        val = -2*np.min(np.min(np.real(np.linalg.eig(rho1)[0])), 0)

        return val
    else:
        return 'NA'

"""
purity(rhog)
Desc: Calculates the purity of the input state.

Parameters
----------
rhog : ndarray
    Density Matrix of the desired state.

Returns
-------
val : float
    The calculated purity ranging from 1/(2^numQubits) to 1.
    A value of one corresponds to a completly pure state.

Other Properties
 -------------- 
entropy;linear_entropy;negativity;concurrence;tangle

See Also
 ------ 
err_functions;getProperties
"""
def purity(rhog):
    return np.real(np.trace(np.dot(rhog, rhog)))

"""
partial_transpose(rhog)
Desc: Returns the partial transpose of the input density matrix. DISCLAIMER : Tests in progress

Parameters
----------
rhog : ndarray
    Input arrays find the partial transpose of.

Returns
-------
rv : ndarray
    Partial transpose of rhog.
"""
def partial_transpose(rhog, n = 0, d = np.nan):
    if min(rhog.shape) == 1:
            rhog = np.dot(rhog, rhog.conj().transpose())

    if d is np.nan:
        n_qubit = int(np.log2(rhog.shape[0]))
        if not n_qubit % 1:
            d = 2*np.ones(n_qubit)
        else:
            print('dimension of rho is incorrect.')

    na = 0
    nb = 0
    nc = 0
    if n < 0:
        na = 1.0
        nb = 1.0
        nc = np.prod(d)
    elif n == 0:
        na = 1.0
        nb = d[n]
        nc = np.prod(d[np.arange(n+1, len(d))])
    elif (n > 0) & (n < len(d)):
        na = np.prod(d[range(n-1)])
        nb = d[n]
        nc = np.prod(d[np.arange(n+1, len(d))])
    elif n == len(d):
        na = np.prod(d[0:n-1])
        nb = d[n]
        nc = 1.0
    elif n > len(d):
        na = np.prod(d)
        nb = 1.0
        nc = 1.0

    if na == 1:
        rv = partial_transpose_helper(rhog, nb)
    # I did't check from here
    else:
        sub_sizes = nb*nc
        y = np.zeros([sub_sizes, sub_sizes, na, na])+0j
        for j in range(sub_sizes):
            for k in range(sub_sizes):
                y[j, k] = rhog[j*sub_sizes:j*sub_sizes+na, k*sub_sizes:k*sub_sizes+na]

        rv = np.zeros([len(rhog), len(rhog)])+0j

        for j in range(na):
            for k in range(na):
                rv[j*nb:j*nb+na, k*nb:k*nb+na] = partial_transpose_helper(y[j, k], nb)

    return rv

"""
performOperation(psi, g)
Desc: Performs the operations on the input State.

Parameters
----------
psi : ndarray
    The input state to do the operation on. Can be a pure state or a density matrix.
g : ndarray with shape = (num operations, 2^numQubits, 2^numQubits)
    The operations you would like to be done. Can be one operation or an array of operations.

Returns
-------
p : ndarray
    The output state after the operations. Will retain the input form.
"""
def performOperation(psi, g):
    p = psi
    if(len(g.shape) == 3):
        # Multiple operations
        for i in range(g.shape[0]):
            p = performOperation(p,g[i])
    else:
        if isStateVector(psi):
            p = ketOperation(p, g)
        else:
            p = densityOperation(p, g)
    return p


"""
random_pure_state(N)
Desc: Returns a random quantum state from a uniform distribution across the space.

Parameters
----------
N : int
    The dimension of the quantum state

Returns
-------
pure_state : ndarray with length = 2^N
    The random quantum state.
"""
def random_pure_state(N = 1):
    pure_state = np.zeros(2**N,dtype=complex)
    for x in range(len(pure_state)):
        pure_state[x] = rand.normal(0, 1)+rand.normal(0, 1)*1j
    length = np.inner(pure_state.conj(),pure_state)
    pure_state = pure_state/np.sqrt(length)
    length = np.inner(pure_state.conj(), pure_state)
    return pure_state


"""
random_density_state(N)
Desc: Returns a random quantum density from an approximate uniform distribution across the space.

Parameters
----------
N :int
    The dimension of the quantum state

Returns
-------
density : ndarray with shape = (2^N, 2^N)
    The random quantum state.
"""
def random_density_state(N=1):
    density = random_ginibre(2 ** N)
    density = np.matmul(density, density.T.conj())
    if np.trace(density).real<10**-6:
        # If trace is close to 0 then resample
        return random_density_state(N)
    else:
        density = density / np.trace(density)
    return density

"""
random_bell_state(N)
Desc: Randomly returns one of the 4 bell state. 
For 1 qubits on of the standard basis states is returned. For states with dimension
greater then 2 the Greenberger-Horne-Zeilingerstate is returned with a random phase.

Parameters
----------
N :int
    The dimension of the quantum state

Returns
-------
pure_state : ndarray with shape = (2^N, 2^N)
    The random quantum state.
"""
def random_bell_state(N=2):
    if N ==1:
        whichState = rand.randint(0,6)
        if whichState == 0:
            return np.array([1,0],dtype=complex)
        elif whichState == 1:
            return np.array([0,1],dtype=complex)
        elif whichState == 2:
            return 1/np.sqrt(2)*np.array([1,1],dtype=complex)
        elif whichState == 3:
            return 1/np.sqrt(2)*np.array([1,-1],dtype=complex)
        elif whichState == 4:
            return 1/np.sqrt(2)*np.array([1,1j],dtype=complex)
        elif whichState == 5:
            return 1/np.sqrt(2)*np.array([1,-1j],dtype=complex)
    if N ==2:
        whichState = rand.randint(0, 4)
        if whichState == 0:
            return 1 / np.sqrt(2) * np.array([1,0,0,1], dtype=complex)
        elif whichState == 1:
            return 1 / np.sqrt(2) * np.array([1,0,0,-1], dtype=complex)
        elif whichState == 2:
            return 1 / np.sqrt(2) * np.array([0,1,1,0], dtype=complex)
        elif whichState == 3:
            return 1 / np.sqrt(2) * np.array([0,1,-1,0], dtype=complex)
    GHZ = np.zeros(2**N,dtype=complex)
    GHZ[0] = 1
    GHZ[-1] = 1
    return 1 / np.sqrt(2)* GHZ


"""
random_ginibre(D)
Desc: Returns a random matrix from the Ginibre ensemble of size DxD. 
This is a complex matrix whos elements are a+ib | a,b iid. Norm(0,1)

Parameters
----------
D :int
    The dimension of the Matrix

Returns
-------
mat : ndarray with shape = (2^N, 2^N)
    The random matrix
"""
def random_ginibre(D=2):
    mat = np.zeros((D, D), dtype=complex)
    for i in range(D):
        for j in range(D):
            mat[i, j] = rand.normal(0, 1) + rand.normal(0, 1) * 1j

    return mat

"""
densityOperation(D)
Desc: Performs the operation on the density matrix

Parameters
----------
psi : ndarray with shape = (2^nQubits, 2^nQubits)
    The state to apply the operation to in density form.
gate : ndarray with shape = (2^nQubits, 2^nQubits)
    The operation to apply to the state.

Returns
-------
psi : ndarray with shape = (2^nQubits, 2^nQubits)
    The state with the operation applied.
"""
def densityOperation(psi, gate):
    return np.matmul(np.matmul(gate, psi), np.conjugate(np.transpose(gate)))

"""
ketOperation(D)
Desc: Performs the operation on the pure state.

Parameters
----------
psi : 1darray with length = 2^nQubits
    The state to apply the operation to in ket form.
gate : ndarray with shape = (2^nQubits, 2^nQubits)
    The operation to apply to the state.

Returns
-------
psi : ndarray with shape = (2^nQubits, 2^nQubits)
    The state with the operation applied.
"""
def ketOperation(psi, gate):
    return np.matmul(gate, psi)

"""
quarterWavePlate(D)
Desc: returns the quantum gate associated with a quarter wave plate.

Parameters
----------
theta : float
    The angle of the wave plate with respect to horizontal.
"""
def quarterWavePlate(theta):
    return np.array([
        [np.cos(theta)**2+1j*np.sin(theta)**2   ,   (1-1j)*np.cos(theta)*np.sin(theta)  ],
        [(1-1j)*np.cos(theta)*np.sin(theta)     ,   np.sin(theta)**2+1j*np.cos(theta)**2]
    ],dtype=complex)

"""
halfWavePlate(D)
Desc: returns the quantum gate associated with a half wave plate.

Parameters
----------
theta : float
    The angle of the wave plate with respect to horizontal.
"""
def halfWavePlate(theta):
    return np.array([
        [np.cos(theta)**2-np.sin(theta)**2      ,   2*np.cos(theta)*np.sin(theta)  ],
        [2*np.cos(theta)*np.sin(theta)          ,   np.sin(theta)**2-np.cos(theta)**2]
    ],dtype=complex)

"""
getWavePlateBasis(theta_qwp,theta_hwp,flipPBS)
Desc: Given the angles for the QWP and HWP plate find the measurement basis. PBS is assumed to transmit 
Horizontally polarized light and reflect vertical. This function does not take into account crosstalk.

Parameters
----------
theta_qwp : string
    The angle with respect to horizontal for the quarter wave plate.
theta_hwp : string
    The angle with respect to horizontal for the quarter wave plate.    
flipPBS: bool
        Set this to true to assume the PBS transmits V and reflects H

Returns
-------
basis : ndarray with shape (2,2)
    Top row : State the original state was projected onto given that it transmitted through the PBS,
    Bottom row : State the original state was projected onto given that it reflected off the PBS
"""
def getWavePlateBasis(theta_qwp,theta_hwp,flipPBS=False):
    basis = np.eye(2,dtype=complex)
    if(flipPBS):
        basis = np.fliplr(basis)
    basis = np.matmul(halfWavePlate(theta_hwp).T.conj(),basis)
    basis = np.matmul(quarterWavePlate(theta_qwp).T.conj(),basis)
    # return the transpose so that basis[0] returns state 1
    # no need to do conj, its just a reordering of the numbers
    return basis.T

"""
removeGlobalPhase(D)
Desc: Factors out the global phase of the given state by dividing the entire state by the phase of the first component.

Parameters
----------
pure_state : 1darray with length = 2^nQubits
    The state in ket form.

Returns
-------
psi : ndarray with shape = (2^nQubits, 2^nQubits)
    The state with global phase factored out.
"""
def removeGlobalPhase(pure_state):
    # Normalize
    norm = np.dot(pure_state.conj(), pure_state)
    pure_state = pure_state / norm
    # get phase of first number
    [magn,phase] = complexToPhaser(pure_state[0])
    complexPhaser = phaserToComplex(np.array([1,phase]))
    # divide state by first phase
    pure_state = pure_state/complexPhaser
    return pure_state


from __future__ import print_function
import argparse
import os
import sys
import warnings
warnings.filterwarnings('ignore')
import numpy as np
from numpy.core.defchararray import add
import QuantumTomography as qKLib

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""

# This script is used to run quantum tomography from the command line.
#
# Running the command Quantum-Tomography from the command line in the package directory will run this script


def file_path(string):
    if os.path.isfile(string) or string == "../../" :
        return string
    else:

        raise OSError(string)

def dir_path(string):
    if os.path.isdir(string) or string == "../../" :
        return string
    else:

        raise OSError(string)

def main():
    # create argument parser object
    parser = argparse.ArgumentParser(description = "Quantum Tomography")

    parser.add_argument("-i", "--eval", type = file_path, nargs = 1,
                        metavar = "evalFile", default = None, help = "The full path to the file that contains the data and configuration for the tomography.")

    parser.add_argument("-s", "--save", type = dir_path, nargs = 1,
                        metavar = "outPutFolder", default = None, help = "The full path to the folder where you want the output to be saved. If not included it will not save your data.")

    parser.add_argument("-p", "--pic", action = 'store_true',default = False,
                        help = "Including this will show images of real and imaginary values of the density matrix. If save is also included pictures will be saved and not shown.")


    # parse the arguments from standard input
    args = parser.parse_args()

    try:
        inPutfilePath = args.eval[0]
    except:
        raise ValueError("input not defined")
    try:
        outPutfilePath = args.save[0]
        save = True
    except:
        save = False
    pictures = args.pic

    t = qKLib.Tomography()

    # import the eval file to import both the config and data
    [rho, intensity, fval] = t.importEval(inPutfilePath)
    qKLib.printLastOutput(t)

    if(save and save != "False"):
        if not os.path.exists(outPutfilePath + '/TomoOutPut'):
            os.makedirs(outPutfilePath + '/TomoOutPut')
        # Prints the data to a html file
        FORREPLACE = '<h1 style = "text-align: center;"> Tomography Results</h1>'
        if(pictures):
            import matplotlib.pyplot as plt
            FORREPLACE = FORREPLACE + '<img src = "rhobarReal.png" style = "float: left;" width = "550" height = "auto">'
            FORREPLACE = FORREPLACE + '<img src = "rhobarImag.png" width = "550" height = "auto"><br>'
            qKLib.saveRhoImages(rho,outPutfilePath + '/TomoOutPut')

        FORREPLACE = FORREPLACE + qKLib.matrixToHTML(rho)

        vals = t.getProperties(rho)
        FORREPLACE = str(FORREPLACE) + qKLib.propertiesToHTML(vals)

        # Print out properties of bellSettings
        bs = ''
        if (t.conf['Bellstate'] != 0):
            vals = t.getBellSettings(rho)
            resolution = (np.pi / 2) / (9 * (5 ** 3))
            bs += '<h3>Bell inequality (S_local_realism <= 2)</h3>'
            bs += '<table style = \"width:60%;margin-top:10px;font-size: 15px;padding-bottom:5px;float:none;\"><tr><td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;">Property</td><td  style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;">Value</td>'
            if (vals.shape[1] > 2):
                bs += '<td  style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;">   Error</td></tr>'
            else:
                bs += '<td  style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;"></td></tr>'
            for v in vals:
                bs += '<tr><td>' + v[0] + '</td><td>' + qKLib.floatToString(v[1]) + ' &deg;</td>'
                if (len(v) > 2):
                    bs += '<td>+/- ' + qKLib.floatToString(v[2]) + '&deg;</td></tr> \n'
                else:
                    bs += "<td></td></tr>"
            bs += '<tr><td>resolution</td><td>' + qKLib.floatToString(resolution * 180 / np.pi) + '&deg;</td><td></tr></tr> \n'
            bs += '</td></tr></table>'
        bs = bs.replace('+/- nan&deg;', 'Error')
        bs = bs.replace('nan &deg;', 'Error')

        # Print out time at the bottom of page
        FORREPLACE = str(FORREPLACE) + '<br><br>\n' + bs
        FORREPLACE += '</font></div>'


        # Update the html file
        fff = "<html><head></head><body>TOREPLACE</body></html>"

        fff = fff.replace('TOREPLACE', str(FORREPLACE))

        with open(outPutfilePath + '/TomoOutPut/outPut.html', 'w') as ff:
            ff.write(fff)
            ff.close()
        if(outPutfilePath[-1] == "\\" or outPutfilePath[-1] == "/"):
            print("Output saved to "+outPutfilePath + 'TomoOutPut')
        else:
            if (outPutfilePath.find("\\") == -1):
                print("Output saved to " + outPutfilePath + '/TomoOutPut')
            else:
                print("Output saved to " + outPutfilePath + '\\TomoOutPut')
    else:
        if(pictures):
            import matplotlib.pyplot as plt
            qKLib.makeRhoImages(rho,plt)
            plt.show()

if __name__ == "__main__":
    main()



try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
import re

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""

class ConfDict(MutableMapping):
    """A dictionary where the casing of the keys don't matter
        and values can be automatically handled."""

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        return self.store[self._keytransform(key)]

    def __setitem__(self, key, value):
        self.store[self._keytransform(key)] = self._valuetransform(value)

    def __delitem__(self, key):
        del self.store[self._keytransform(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def _keytransform(self, key):
        return key.lower()
    def _valuetransform(self, value):
        if (isinstance(value, str)):
            if (value.lower() == "yes" or
                value.lower() == "true" or
                value.lower() == "t" or
                value.lower() == "y"):
                value = 1
            elif (value.lower() == "no" or
                  value.lower() == "false" or
                  value.lower() == "f"or
                  value.lower() == "f"):
                value = 0
            elif(value.upper() == "LINEAR" or
                value.upper() == "MLE" or
                value.upper() == "HMLE" or
                value.upper() == "BME"):
                return value.upper()
            else:
                raise ValueError('Invalid Conf Setting of "' + value+'"')

        return value



def getValidFileName(fileName):
    newFileName = re.sub(r"^\\.+", "", fileName)
    newFileName = re.sub(r"[\\\\/:*?\"<>|]", "", newFileName)
    if newFileName == "":
        raise ValueError("File Name : '" + fileName + "' results in an empty fileName!");
    return newFileName


from __future__ import print_function
from .TomoFunctions import *
from .TomoDisplay import floatToString
from .TomoClassHelpers import *
from .Utilities import ConfDict,getValidFileName
import numpy as np
from scipy.optimize import leastsq,minimize
import warnings

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""

"""
    class Tomography()
    Desc: This is the main tomography object that the library is built around. If you are looking to perform multiple tomographies
          it is advised to create multiple Tomography objects.

    See Also
     ------ 
    importEval;importConf;importData
    """
class Tomography():

    # # # # # # # # # # # #
    '''Public Variables'''
    # # # # # # # # # # # #

    # conf['NQubits']: >= 1, it will take a much longer time for more qubits.
    # conf['NDetectors']: 1 or 2
    # conf['Crosstalk']: [[C0->0, C0->1], [C1->0, C1->1]]
    # conf['Bellstate']: 'no' or 'yes'
    # conf['DoDriftCorrection'] = 'no' or 'yes'
    # conf['DoAccidentalCorrection'] = 'no' or 'yes'
    # conf['DoErrorEstimation']: >= 0
    # conf['Window']: 0 or array like, dimension = 1
    # conf['Efficiency']: 0 or array like, dimension = 1
    # conf['Beta']: 0 to 0.5, depending on purity of state and total number of measurements.
    # conf['ftol']: Relative error desired in the sum of squares. Passed directly to scipy.optimize.leastsq
    # conf['xtol']: Relative error desired in the approximate solution. Passed directly to scipy.optimize.leastsq
    # conf['gtol']: Orthogonality desired between the function vector and the columns of the Jacobian. Passed directly to scipy.optimize.leastsq
    # conf['maxfev']: The maximum number of calls to the function. Passed directly to scipy.optimize.leastsq

    # tomo_input: array like, dimension = 2.
    # input data of the last tomography run.

    # For n detectors:
    # tomo_input[:, 0]: times
    # tomo_input[:, np.arange(1, n_qubit+1)]: singles
    # tomo_input[:, n_qubit+1]: coincidences
    # tomo_input[:, np.arange(n_qubit+2, 3*n_qubit+2)]: measurements
    #
    # For 2n detectors:
    # tomo_input[:, 0]: times
    # tomo_input[:, np.arange(1, 2*n_qubit+1)]: singles
    # tomo_input[:, np.arange(2*n_qubit+1, 2**n_qubit+2*n_qubit+1)]: coincidences
    # tomo_input[:, np.arange(2**n_qubit+2*n_qubit+1, 2**n_qubit+4*n_qubit+1)]: measurements

    # last_rho: The predicted density matrix of the last tomography run.
    # last_intensity: The predicted intensity of the state for the last tomography run.
    # last_fval: The final value of the internal optimization function for the last tomography run.
    # last_input: The last tomo_input matrix used
    # mont_carlo_states: The generated monte carlo states.
    # intensity: Relative pump power (arb. units) during measurement for the last tomography run.
    # err_functions: These are the properties that you want to be calculated when getProperties is called.

    # # # # # # # # # #
    '''Constructors'''
    # # # # # # # # # #

    """
    Default Constructor
    Desc: This initializes a default tomography object
    """
    def __init__(self,nQ = 2):
        self.conf = ConfDict([('NQubits',nQ), ('NDetectors', 1),
                              ('Crosstalk', np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])),
                              ('Bellstate', 0),('DoDriftCorrection', 0), ('DoAccidentalCorrection', 0),
                              ('DoErrorEstimation', 0),('Window', [1]),('Efficiency', [1]),('RhoStart', []),
                              ('Beta', 0),('ftol', 1.49012e-08),('xtol', 1.49012e-08),('gtol', 0.0),('maxfev', 0)])
        self.err_functions = ['concurrence', 'tangle', 'entropy', 'linear_entropy', 'negativity', 'purity']
        self.mont_carlo_states = list()
    """
    setConfSetting(setting, val)
    Desc: Sets a specific self.conf setting. This is no longer needed but will be kept in the class to
    support old code.

    Parameters
    ----------
    setting : string
        The setting you want to update.
        Possible values are ['NQubits', 'NDetectors', 'Crosstalk', 'Bellstate', 'DoDriftCorrection', 
        'DoAccidentalCorrection', 'DoErrorEstimation', 'Window', 'Efficiency', 'RhoStart', 'Beta','ftol','xtol','gtol'
        'maxfev']
    val : ndarray, int, or string
            The new value you want to the setting to be.
    """
    def setConfSetting(self, setting, val):
        warnings.warn('As of v1.0.3.7 setConfSetting() is no longer needed to set a conf setting. '
                                 'Settings can now be set directly like a normal dictionary. ex: tomo.conf["DoDriftCorrection"] = 1', DeprecationWarning)
        if (isinstance(val, str)):
            if (val.lower() == "yes" or val.lower() == "true"):
                valC = 1
            elif (val.lower() == "no" or val.lower() == "false"):
                valC = 0
        self.conf[setting] = valC

    """
    importConf(conftxt)
    Desc: Import a text file containing the configuration settings.

    Parameters
    ----------
    conftxt : string
        path to configuration file
    """
    def importConf(self, conftxt):
        conf = self.conf
        exec(compile(open(conftxt, "rb").read(), conftxt, 'exec'))

    """
    importData(datatxt)
    Desc: Import a text file containing the tomography data and run tomography.

    Parameters
    ----------
    datatxt : string
       path to data file

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : float
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def importData(self, datatxt):
        exec(compile(open(datatxt, "rb").read(), datatxt, 'exec'))
        return self.StateTomography_Matrix(locals().get('tomo_input'), locals().get('intensity'),method=self.conf["method"])

    """
    importEval(evaltxt)
    Desc: Import a eval file containing the tomography data and the configuration setting, then run tomography.

    Parameters
    ----------
    evaltxt : string
        path to eval file

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : float
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def importEval(self, evaltxt):
        conf = self.conf
        exec(compile(open(evaltxt, "rb").read(), evaltxt, 'exec'))
        if "method" in self.conf.keys():
            return self.StateTomography_Matrix(locals().get('tomo_input'), locals().get('intensity'),method=self.conf["method"])
        else:
            return self.StateTomography_Matrix(locals().get('tomo_input'), locals().get('intensity'))

    # # # # # # # # # # # # # #
    '''Tomography Functions'''
    # # # # # # # # # # # # # #


    """
    StateTomography(measurements, counts)
    Desc: Main function that runs tomography. This function requires a set of measurements and a set of counts.

    Parameters
    ----------
    measurements : ndarray shape = ( Number of measurements , 2*NQubits )
        Each row in the matrix is a set of independent measurements.
    counts : ndarray shape = (Number of measurements, NDetectors**NQubits )
        Each row in the matrix is a set of independent measurements.
    crosstalk : ndarray shape = ( Number of measurements , 2**NQubits,2**NQubits ) (optional)
        The crosstalk matrix to compensate for inefficentcies in your beam splitter.
    efficiency : 1darray with length = NQubits*NDetectors (optional)
        The relative efficienies between your detector pairs.
    time :  1darray with length = Number of measurements (optional)
        The total duration of each measurment.
    singles : ndarray shape = ( Number of measurements , 2*NQubits ) (optional)
        The singles counts on each detector.
    window : 1darray with length = NQubits*NDetectors (optional)
        The coincidence window duration for each detector pair.
    error : int (optional)
        The number of monte carlo states used to estimate the properties of the state.
    intensities : 1darray with length = Number of measurements (optional)
        The relative intensity of each measurement. Used for drift correction.
    method : ['MLE','HMLE','BME','LINER'] (optional)
        Which method to use to run tomography. Default is MLE

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : The predicted overall intensity used to normalize the state.
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def StateTomography(self, measurements, counts, crosstalk=-1, efficiency=0, time=-1, singles=-1, window=0, error=0,
                        intensities=-1, method="MLE"):
        tomo_input = self.buildTomoInput(measurements, counts, crosstalk, efficiency, time, singles, window, error)
        return self.StateTomography_Matrix(tomo_input, intensities, method=method)

    """
    StateTomography_Matrix(tomo_input, intensities)
    Desc: Main function that runs tomography.

    Parameters
    ----------
    tomo_input : ndarray
        The input data for the current tomography. Example can be seen at top of page.
    intensities : 1darray with length = number of measurements (optional)
        Relative pump power (arb. units) during measurement; used for drift correction. Default will be an array of ones
    method : ['MLE','HMLE','BME','LINER'] (optional)
            Which method to use to run tomography. Default is MLE

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : The predicted overall intensity used to normalize the state.
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def StateTomography_Matrix(self, tomo_input, intensities = -1,method="MLE",_saveState=True):
        # define a uniform intensity if not stated
        if (isinstance(intensities, int)):
            self.intensities = np.ones(tomo_input.shape[0])
        elif(len(intensities.shape) == 1 and intensities.shape[0] == tomo_input.shape[0]):
            self.intensities = intensities
        else:
            raise ValueError("Invalid intensities array")
        if any(self.intensities-np.ones_like(self.intensities))>10**-6:
            self.conf['DoDriftCorrection'] = 1

        # filter the data
        self.last_input = tomo_input
        self.conf['Method'] = method
        [coincidences, measurements_densities, measurements_pures, accidentals,overall_norms] = self.filter_data(tomo_input)

        # get the starting state from tomography_LINEAR if not defined
        starting_matrix = self.conf['RhoStart']
        if method.upper() == "LINEAR" or not isinstance(starting_matrix,np.ndarray):
            try:
                # todo: go over linear tomography. Clean it up. Figure out why it fails on the very rare occasion.
                [starting_matrix,inten_linear] = self.tomography_LINEAR(coincidences, measurements_pures,overall_norms)
                # Currently linear tomography gets the phase wrong. So a temporary fix is to just transpose it.
                starting_matrix = starting_matrix.transpose()
                starting_matrix = make_positive(starting_matrix)
                starting_matrix = starting_matrix / np.trace(starting_matrix)
            except:
                raise RuntimeError('Failed to run linear Tomography')


        # Run tomography and find an estimate for the state
        if method == "MLE":
            [rhog, intensity, fvalp] = self.tomography_MLE(starting_matrix, coincidences, measurements_densities, accidentals,overall_norms)
        elif method.upper() == "HMLE":
            [rhog, intensity, fvalp] = self.tomography_HMLE(starting_matrix, coincidences, measurements_densities,
                                                           accidentals, overall_norms)
        # elif method.upper() == "BME":
        #     [rhog, intensity, fvalp] = self.tomography_BME(starting_matrix, coincidences, measurements_densities,accidentals,overall_norms)
        elif method.upper() == "LINEAR":
            [rhog, intensity, fvalp] = [starting_matrix,inten_linear,0]
        else:
            raise ValueError("Invalid Method name: " + str(method))

        if _saveState:
            # save the results
            self.last_rho = rhog.copy()
            self.last_intensity = intensity
            self.last_fval = fvalp
            # Reset the monte carlo states after doing a tomography
            self.mont_carlo_states = list([[rhog, intensity, fvalp]])

        return [rhog, intensity, fvalp]


    # This is a temporary function. It's here in case there is old code that still uses state_tomography.
    # Eventually this should be removed in a later version
    def state_tomography(self, tomo_input, intensities = -1,method="MLE",_saveState=True):
        warnings.warn('state_tomography will be removed in a future version. It is replaced by StateTomography_Matrix which does the same exact thing.', DeprecationWarning)
        return self.StateTomography_Matrix(tomo_input,intensities,method,_saveState)

    """
    tomography_MLE(starting_matrix, coincidences, measurements, accidentals,overall_norms)
    Desc: Runs tomography using maximum likelyhood estimation.

    Parameters
    ----------
    starting_matrix : ndarray with shape = (2^numQubits, 2^numQubits)
        The starting predicted state.
    coincidences : ndarray shape = (Number of measurements, NDetectors**NQubits)
        The counts of the tomography.
    measurements_densities : ndarray with shape = (NDetectors*number of measurements,2^numQubits, 2^numQubits) 
        The measurements of the tomography in density matrix form.
    accidentals : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The accidental values of the tomography. Used for accidental correction.
    overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The relative weights of each measurment. Used for drift correction.

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : float
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def tomography_MLE(self, starting_matrix, coincidences, measurements, accidentals,overall_norms=-1):
        # If overall_norms not given then assume uniform
        if not isinstance(overall_norms,np.ndarray):
            overall_norms = np.ones(coincidences.shape[0])
        elif not (len(overall_norms.shape) == 1 and overall_norms.shape[0] == coincidences.shape[0]):
            raise ValueError("Invalid intensities array")

        init_intensity = np.mean(np.multiply(coincidences, 1/overall_norms)) * (starting_matrix.shape[0])
        starting_tvals = density2t(starting_matrix)
        starting_tvals = starting_tvals + 0.0001
        starting_tvals = starting_tvals * np.sqrt(init_intensity)

        coincidences = np.real(coincidences)
        coincidences = coincidences.flatten()

        final_tvals = leastsq(maxlike_fitness, np.real(starting_tvals),
                              args = (coincidences, accidentals, measurements, overall_norms),
                              ftol=self.conf["ftol"],
                              xtol=self.conf["xtol"],
                              gtol=self.conf["gtol"],
                              maxfev=self.conf["maxfev"])[0]
        fvalp = np.sum(maxlike_fitness(final_tvals, coincidences, accidentals, measurements, overall_norms) ** 2)

        final_matrix = t_to_density(final_tvals, normalize=False)
        intensity = np.trace(final_matrix)
        final_matrix = final_matrix / np.trace(final_matrix)

        intensity = np.float64(np.real(intensity))

        return [final_matrix, intensity, fvalp]

    """
    tomography_HMLE(starting_matrix, coincidences, measurements, accidentals,overall_norms)
    Desc: Runs tomography using hedged maximum likelyhood estimation.

    Parameters
    ----------
    starting_matrix : ndarray with shape = (2^numQubits, 2^numQubits)
        The starting predicted state.
    coincidences : ndarray shape = (Number of measurements, NDetectors**NQubits)
        The counts of the tomography.
    measurements_densities : ndarray with shape = (NDetectors*number of measurements,2^numQubits, 2^numQubits) 
        The measurements of the tomography in density matrix form.
    accidentals : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The accidental values of the tomography. Used for accidental correction.
    overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The relative weights of each measurment. Used for drift correction.

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : float
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    """
    def tomography_HMLE(self, starting_matrix, coincidences, measurements, accidentals,overall_norms=-1):
        # If overall_norms not given then assume uniform
        if not isinstance(overall_norms, np.ndarray):
            overall_norms = np.ones(coincidences.shape[0])
        elif not (len(overall_norms.shape) == 1 and overall_norms.shape[0] == coincidences.shape[0]):
            raise ValueError("Invalid intensities array")

        init_intensity = np.mean(np.multiply(coincidences, 1 / overall_norms)) * (starting_matrix.shape[0])
        starting_tvals = density2t(starting_matrix)
        starting_tvals = starting_tvals + 0.0001
        starting_tvals = starting_tvals * np.sqrt(init_intensity)

        coincidences = np.real(coincidences)
        coincidences = coincidences.flatten()

        bet = self.conf['Beta']
        if bet > 0:
            final_tvals = leastsq(maxlike_fitness_hedged, np.real(starting_tvals),
                                  args=(coincidences, accidentals, measurements, bet, overall_norms),
                                  ftol=self.conf["ftol"],
                                  xtol=self.conf["xtol"],
                                  gtol=self.conf["gtol"],
                                  maxfev=self.conf["maxfev"])[0]
            fvalp = np.sum(maxlike_fitness_hedged(final_tvals, coincidences, accidentals, measurements, bet,
                                                      overall_norms) ** 2)
        else:
            raise ValueError("To use Hedged Maximum Likelihood, Beta must be a positive number.")

        final_matrix = t_to_density(final_tvals,normalize=False)
        intensity = np.trace(final_matrix)
        final_matrix = final_matrix / np.trace(final_matrix)

        intensity = np.float64(np.real(intensity))

        return [final_matrix, intensity, fvalp]

    # todo: change these back to double quotes when it is ready to be displayed on documentation page.
    '''
    tomography_BME(starting_matrix, coincidences, measurements, accidentals,overall_norms)
    Desc: Runs tomography using bayesian mean estimation. Currently in progress.

    Parameters
    ----------
    starting_matrix : ndarray with shape = (2^numQubits, 2^numQubits)
        The starting predicted state.
    coincidences : ndarray shape = (Number of measurements, NDetectors**NQubits)
        The counts of the tomography.
    measurements : ndarray with shape = (NDetectors*number of measurements,2^numQubits, 2^numQubits) 
        The measurements of the tomography in density matrix form.
    accidentals : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The accidental values of the tomography. Used for accidental correction.
    overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The relative weights of each measurment. Used for drift correction.

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The predicted density matrix.
    intensity : float
        The predicted overall intensity used to normalize the state.
    fvalp : float
        Final value of the internal optimization function. Values greater than the number
        of measurements indicate poor agreement with a quantum state.
    '''
    # def tomography_BME(self,starting_matrix, coincidences, measurements, accidentals,overall_norms):
    #
    #     # algorithm parameters
    #     preallocationSize = 50000  # preallocate memory for speed
    #     changeThreshold = 10 ** -19 # stopping condition threshold
    #     numberOfPriorSamples = 10000  # min number of samples to sample from the inital prior
    #     updateFrequency = 300  # controls how often the estimate of the posterior is updated, and how often stability is checked
    #     max_iterations = 98  # The maximum number of iterations(updates)
    #
    #     # initialize
    #     d = starting_matrix.shape[0]
    #     i = 0
    #     iterationCounter = 0
    #     eps = 2.2204e-16
    #     prevMean_rho = np.zeros((d,d))
    #     prev_minIndex = -1
    #     mean_tvals = -1
    #     cov_tvals = -1
    #
    #     # Todo TEMP
    #     tempLogLikeMean = np.zeros(101)
    #     tempLogLikeMax = np.zeros(101)
    #     tempLogLikeSD = np.zeros(101)
    #
    #
    #     # Lists that are evaluated periodically
    #     stabilityHistory = np.zeros(max_iterations*3)
    #
    #     # List of Sample States
    #     sampleStates_rho = np.zeros((d,d,preallocationSize), dtype=complex)
    #     sampleStates_loglike = np.zeros(preallocationSize)
    #     sampleStates_tvals = np.zeros((preallocationSize, d ** 2))
    #
    #     # estimate the counts intensity
    #     norm = sum(coincidences / (len(coincidences) * d))
    #     minimization_output = minimize(log_likelyhood, norm,
    #                           args=(density2t(starting_matrix), coincidences, measurements, accidentals,overall_norms))
    #     norm = minimization_output.x[0]
    #     base_loglike = minimization_output.fun
    #
    #     # SAMPLE FROM PRIOR
    #     # Samples are drawn from the ginibre distribution
    #     stillUsingPrior = 1
    #     while stillUsingPrior:
    #         # expand memory in chunks if preallocation size is exceeded
    #         if (i >= len(sampleStates_loglike)):
    #             sampleStates_rho = np.concatenate((sampleStates_rho, np.zeros((d,d,i))), axis=2)
    #             sampleStates_loglike = np.concatenate((sampleStates_loglike, np.zeros(i)))
    #             sampleStates_tvals = np.concatenate((sampleStates_tvals, np.zeros((i, d ** 2))), axis=0)
    #
    #         # Sample from the ginibre distribution
    #         random_rho = random_density_state(self.conf['NQubits'])
    #         random_tvals = density2t(random_rho)
    #         random_loglike = log_likelyhood(norm, random_rho, coincidences, measurements, accidentals,overall_norms)
    #
    #         # estimate the counts intensity
    #         if random_loglike < base_loglike:
    #             minimization_output = minimize(log_likelyhood, norm,
    #                                            args=(random_rho, coincidences, measurements,
    #                                                  accidentals,
    #                                                  overall_norms))
    #             norm = minimization_output.x[0]
    #             random_loglike = minimization_output.fun
    #             base_loglike = random_loglike
    #
    #         # Store sample state, tval, and its loglikelyhood
    #         sampleStates_rho[:,:,i] = random_rho
    #         sampleStates_tvals[i] = random_tvals
    #         sampleStates_loglike[i] = random_loglike
    #
    #
    #         # Periodically perform the following tasks
    #         if i % updateFrequency == 0 and i !=0:
    #             # Normalize the likelihoods
    #             [sampleStates_normlike, minIndex,sampleStates_scaledloglike] = normalizeLikelihoods(sampleStates_loglike[:i + 1])
    #             unNormedMean_rho = np.dot(sampleStates_rho[:, :, :i + 1], np.exp(-1*sampleStates_scaledloglike))
    #
    #             # Update the mean
    #             tr = np.trace(unNormedMean_rho)
    #             if tr == 0:
    #                 mean_rho = np.zeros_like(prevMean_rho)
    #                 stabilityHistory[iterationCounter] = float('inf')
    #             else:
    #                 mean_rho = unNormedMean_rho / tr
    #                 if iterationCounter != 0:
    #                     # Calculate the stability
    #                     sectionLikelihood = sum(np.exp(-1 * sampleStates_scaledloglike[(i + 1 - updateFrequency):(i + 1)]))
    #                     totalLikelihood = sum(np.exp(-1 * sampleStates_scaledloglike[1:(i + 1)]))
    #                     normFactor = totalLikelihood/((iterationCounter+1)*sectionLikelihood)
    #                     infidelity = (1 - np.real(fidelity(mean_rho, prevMean_rho)))
    #                     # ensure rounding never leads the infidelity to be negative
    #                     if infidelity < 0:
    #                         infidelity = eps
    #                     stabilityHistory[iterationCounter] = infidelity * normFactor
    #                 else:
    #                     stabilityHistory[iterationCounter] = 0
    #             prevMean_rho = mean_rho
    #
    #             # If we have met the min required amount of prior samples try to switching to the posterior estimate
    #             if i >= numberOfPriorSamples:
    #                 contributingParams = sampleStates_normlike > 0
    #                 firstMax_like = max(sampleStates_normlike)
    #                 secondMax_like = max(sampleStates_normlike[sampleStates_normlike < firstMax_like])
    #
    #                 # IMPOSE STOPPING CONDITION IF TOO MANY ITERATIONS REACHED
    #                 if (iterationCounter > max_iterations):
    #                     warnings.warn(
    #                         "MAX ITER REACHED - BME (Prior). Stability: " + str(
    #                             stabilityHistory[iterationCounter]))
    #                     try:
    #                         [mean_tvals, cov_tvals] = weightedcov(sampleStates_tvals[:i + 1][contributingParams, :],
    #                                                               sampleStates_normlike[contributingParams])
    #                     except:
    #                         pass
    #                     stillUsingPrior = 0
    #                     i += 1
    #                     iterationCounter = iterationCounter + 1
    #                     break
    #
    #                 # Switch to the posterior under the right conditions
    #                 # 1.) We must ensure one of our samples does not make up majority of the contribution in estimating
    #                 #     the posterior parameters
    #                 # 2.) Cov must be positive semi - definite
    #                 if (secondMax_like / firstMax_like) >= .01:
    #                     [mean_tvals, cov_tvals] = weightedcov(sampleStates_tvals[:i + 1][contributingParams, :],
    #                                                           sampleStates_normlike[contributingParams])
    #                     # Try to find eigenvals. If it can't then keep sampling.
    #                     try:
    #                         if all(np.linalg.eigvals(cov_tvals) >= 0):
    #                             # Stop using the prior
    #                             stillUsingPrior = 0
    #                             i += 1
    #                             iterationCounter = iterationCounter + 1
    #                             break
    #                     except:
    #                         pass
    #             # Increase iteration number
    #             iterationCounter = iterationCounter + 1
    #         i = i + 1
    #
    #
    #     iterPrior = iterationCounter
    #     self.priorsToConverge = i-1
    #
    #
    #     # todo TEMP
    #     tempLogLikeMean[0] = np.average(normalizeLikelihoods(sampleStates_loglike[:i])[0])
    #     tempLogLikeMax[0] = np.max(normalizeLikelihoods(sampleStates_loglike[:i])[0])
    #     tempLogLikeSD[0] = np.std(normalizeLikelihoods(sampleStates_loglike[:i])[0])
    #
    #
    #     # SAMPLE FROM POSTERIOR
    #     # Samples by drawing tVals from a multivariate normal distro
    #     stillUsingPosterior = 1
    #     while stillUsingPosterior:
    #         # expand memory in chunks if preallocation size is exceeded
    #         if (i >= len(sampleStates_loglike)):
    #             sampleStates_rho = np.concatenate((sampleStates_rho, np.zeros((d,d,i))), axis=2)
    #             sampleStates_loglike = np.concatenate((sampleStates_loglike, np.zeros(i)))
    #             sampleStates_tvals = np.concatenate((sampleStates_tvals, np.zeros((i, d ** 2))), axis=0)
    #
    #         # Draws by sampling the tVals from a multivariate normal distro
    #         random_tvals = rand.multivariate_normal(mean_tvals, cov_tvals)
    #         random_rho = t_to_density(random_tvals)
    #         random_loglike = log_likelyhood(norm, random_rho, coincidences, measurements, accidentals,overall_norms)
    #
    #         # estimate the counts intensity
    #         if random_loglike < base_loglike:
    #             minimization_output = minimize(log_likelyhood, norm,
    #                                            args=(random_rho, coincidences, measurements,
    #                                                  accidentals,
    #                                                  overall_norms))
    #             norm = minimization_output.x[0]
    #             random_loglike = minimization_output.fun
    #             base_loglike = random_loglike
    #
    #         # Store sample state, tval, and its loglikelyhood
    #         sampleStates_rho[:, :, i] = random_rho
    #         sampleStates_tvals[i] = random_tvals
    #         sampleStates_loglike[i] = random_loglike
    #
    #         # Periodically perform the following tasks
    #         if i % updateFrequency == 0:
    #             # Normalize the likelihoods
    #             [sampleStates_normlike, minIndex, sampleStates_scaledloglike] = normalizeLikelihoods(
    #                 sampleStates_loglike[:i + 1])
    #             unNormedMean_rho = np.dot(sampleStates_rho[:, :, :i + 1], np.exp(-1*sampleStates_scaledloglike))
    #
    #
    #             # todo TEMP
    #             tempLogLikeMean[iterationCounter-iterPrior+1] = np.average(sampleStates_normlike)
    #             tempLogLikeMax[iterationCounter-iterPrior+1] = max(sampleStates_normlike)
    #             tempLogLikeSD[iterationCounter-iterPrior+1] = np.std(sampleStates_normlike)
    #
    #
    #
    #             # Update the mean
    #             tr = np.trace(unNormedMean_rho)
    #             if tr == 0:
    #                 mean_rho = np.zeros_like(prevMean_rho)
    #                 stabilityHistory[iterationCounter] = float('inf')
    #             else:
    #                 mean_rho = unNormedMean_rho / tr
    #
    #                 # Calculate the stability
    #                 sectionLikelihood = sum(np.exp(-1*sampleStates_scaledloglike[(i + 1 - updateFrequency):(i + 1)]))
    #                 totalLikelihood = sum(np.exp(-1*sampleStates_scaledloglike[1:(i + 1)]))
    #                 normFactor = totalLikelihood / ((iterationCounter + 1) * sectionLikelihood)
    #                 infidelity = (1 - np.real(fidelity(mean_rho, prevMean_rho)))
    #                 # ensure rounding never leads the infidelity to be negative
    #                 if infidelity < 0:
    #                     infidelity = eps
    #                 stabilityHistory[iterationCounter] = infidelity * normFactor
    #
    #                 # Check if the stopping condition is met
    #                 if (iterationCounter > iterPrior) and (stabilityHistory[iterationCounter] < changeThreshold):
    #                     stillUsingPosterior = 0
    #
    #             prevMean_rho = mean_rho
    #
    #             # IMPOSE STOPPING CONDITION IF TOO MANY ITERATIONS REACHED
    #             if (iterationCounter > (max_iterations+iterPrior)):
    #                 warnings.warn(
    #                     "MAX ITER REACHED - BME (Post). Stability: " + str(stabilityHistory[iterationCounter]))
    #                 stillUsingPosterior = 0
    #
    #             contributingParams = sampleStates_normlike > 0
    #             firstMax_like = max(sampleStates_normlike)
    #             secondMax_like = max(sampleStates_normlike[sampleStates_normlike < firstMax_like])
    #
    #             # Update the posterior estimate only if:
    #             # 1.) We must ensure one of our samples make up majority of the contribution in estimating
    #             #     the posterior parameters
    #             # 2.) Cov must be positive semi - definite
    #             if (secondMax_like / firstMax_like) > .001:
    #                 [newMean_tvals, newCov_tvals] = weightedcov(sampleStates_tvals[:i + 1][contributingParams, :],
    #                                                             sampleStates_normlike[contributingParams])
    #                 try:
    #                     # Try to find eigenvals. If it can't then keep sampling.
    #                     if all(np.linalg.eigvals(newCov_tvals) >= 0):
    #                         cov_tvals = newCov_tvals
    #                         mean_tvals = newMean_tvals
    #                 except:
    #                     pass
    #             # Increase iteration number
    #             iterationCounter = iterationCounter + 1
    #         # Increase sample number
    #         i = i + 1
    #     self.stabilityHistory = stabilityHistory
    #     print("done.",end="")
    #     samplesToSC = i
    #
    #
    #     self.tempData = [tempLogLikeMean,tempLogLikeSD,tempLogLikeMax]
    #     return [mean_rho, norm, samplesToSC]

    """
    tomography_LINEAR(coincidences, measurements,overall_norms)
    Desc: Uses linear techniques to find a starting state for maximum likelihood estimation.

    Parameters
    ----------
    coincidences : ndarray shape = (Number of measurements, NDetectors**NQubits)
        The counts of the tomography.
    measurements : ndarray with shape = (NDetectors*number of measurements,2^numQubits) 
        The measurements of the tomography in pure state form.
    overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The relative weights of each measurment. Used for drift correction.

    Returns
    -------
    rhog : ndarray with shape = (2^numQubits, 2^numQubits)
        The starting predicted state.
    intensity : float
        The predicted overall intensity used to normalize the state.
    """
    def tomography_LINEAR(self, coincidences, measurements, overall_norms=-1,m_set = ()):
        # If overall_norms not given then assume uniform
        if not isinstance(overall_norms,np.ndarray):
            overall_norms = np.ones(coincidences.shape[0])
        elif not (len(overall_norms.shape) == 1 and overall_norms.shape[0] == coincidences.shape[0]):
            raise ValueError("Invalid intensities array")

        if m_set == ():
            m_set = independent_set(measurements)
        if np.isscalar(m_set):
            n = len(coincidences)
            linear_measurements = measurements
            linear_data = coincidences
        else:
            n = int(np.sum(m_set))
            linear_measurements = measurements[(np.rot90(m_set == 1.0)[0])]
            linear_data = coincidences[(np.rot90(m_set == 1.0)[0])]

        linear_rhog = np.zeros([measurements.shape[1], measurements.shape[1]])

        b = b_matrix(linear_measurements)
        b_inv = np.linalg.inv(b)

        m = np.zeros([measurements.shape[1], measurements.shape[1], n]) + 0j
        for j in range(n):
            m[:, :, j] = m_matrix(j, linear_measurements, b_inv)
            linear_rhog = linear_rhog + linear_data[j] * m[:, :, j]

        intensity = np.trace(linear_rhog)
        rhog = linear_rhog / intensity

        return [rhog, intensity]


    """
    filter_data(tomo_input)
    Desc: Filters the data into separate arrays.

    Parameters
    ----------
    tomo_input : ndarray
        The input data for the current tomography. Example can be seen at top of page.

    Returns
    -------
    coincidences : ndarray shape = (Number of measurements, NDetectors**NQubits)
        The counts of the tomography.
    measurements_densities : ndarray with shape = (NDetectors*number of measurements,2^numQubits, 2^numQubits) 
        The measurements of the tomography in density matrix form.
    measurements_pures : ndarray with shape = (NDetectors*number of measurements, 2^numQubits) 
        The measurements of the tomography in pure state form.
    acc : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The accidental values of the tomography. Used for accidental correction.
    overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
        The relative weights of each measurment. Used for drift correction.
    """
    def filter_data(self, tomo_input):
        # Check this settings
        self.checkForInvalidSettings()

        # getting variables
        nbits = self.conf['NQubits']
        times = self.getTimes()
        sings = self.getSingles()
        coinc = self.getCoincidences()
        n_coinc = self.getNumCoinc()
        window = self.conf['Window']
        eff = self.conf['Efficiency']
        crosstalk = self.conf['Crosstalk']
        overall_norms = np.kron(self.intensities, eff)


        # Accidental Correction
        acc = np.zeros_like(coinc)
        if(self.conf['DoAccidentalCorrection'] == 1):
            scalerIndex = np.concatenate((np.ones(nbits - 2), [2, 2]))
            additiveIndex = np.array([0, 1])
            for j in range(2, nbits):
                additiveIndex = np.concatenate(([2*j], additiveIndex))
            if (len(coinc.shape) == 1):
                acc = acc[:, np.newaxis]
            for j in range(n_coinc):
                index = bin(j).split("b")[1]
                index = "0" * (nbits - len(index)) + index
                index = [int(char) for char in index]
                index = index*scalerIndex + additiveIndex
                index = np.array(index, dtype = int)
                acc[:, j] = np.prod(np.real(sings[:, tuple(index)]), axis = 1) * (window[j] * 1e-9 / np.real(times)) ** (nbits - 1)
            if (acc.shape != coinc.shape):
                acc = acc[:, 0]

        # Get measurements
        measurements_raw = self.getMeasurements()
        measurements_pures = np.zeros([np.prod(coinc.shape), 2 ** nbits],dtype=complex)
        # NEW WAY
        measurements_densities = np.zeros(
            (tomo_input.shape[0], self.getNumCoinc(), 2 ** nbits, 2 ** nbits), dtype=complex)
        for j in range(tomo_input.shape[0]):
            meas_basis_densities = np.zeros([2 ** nbits, 2 ** nbits, 2 ** nbits]) + 0j
            meas_basis_pures = 1
            for k in range(nbits):
                alpha = measurements_raw[j][2 * k]
                beta = measurements_raw[j][2 * k + 1]
                psi_transmit = np.array([alpha, beta])
                psi_reflect = np.array([np.conj(beta), np.conj(-alpha)])
                meas_pure = np.outer((np.array([1, 0])), psi_transmit) + np.outer((np.array([0, 1])), psi_reflect)
                # Changed from tensor_product to np.kron
                meas_basis_pures = np.kron(meas_basis_pures, meas_pure)
            for k in range(2 ** nbits):
                meas_basis_densities[k, :, :] = np.outer(meas_basis_pures[:, k].conj().transpose(),
                                                         meas_basis_pures[:, k])
            for k in range(self.getNumCoinc()):
                for l in range(2 ** nbits):
                    measurements_pures[j * n_coinc + k, :] = meas_basis_pures[:, k].conj().transpose()
                    measurements_densities[j, k, :, :] = measurements_densities[j, k, :, :] + meas_basis_densities[l,:,:] \
                                                                                                * crosstalk[k, l]
        # Flatten the arrays
        measurements_densities = measurements_densities.reshape((np.prod(measurements_densities.shape[:2]),
                                                                    measurements_densities.shape[2],
                                                                    measurements_densities.shape[3]))

        # Normalize measurements
        for j in range(measurements_pures.shape[0]):
            norm = np.dot(measurements_pures[j].conj(),measurements_pures[j])
            measurements_pures[j] = measurements_pures[j]/np.sqrt(norm)

        for j in range(measurements_densities.shape[0]):
            norm = np.trace(measurements_densities[j])
            measurements_densities[j] = measurements_densities[j]/norm


        # Flatten the arrays
        coincidences = coinc.reshape((np.prod(coinc.shape), 1))
        acc = acc.reshape((np.prod(acc.shape), 1))
        return [coincidences, measurements_densities, measurements_pures, acc,overall_norms]

    """
    buildTomoInput(tomo_input, intensities)
    desc : This function build an input matrix based on a variety of inputs.

    Parameters
    ----------
    measurements : ndarray shape = ( Number of measurements , 2*NQubits )
        Each row in the matrix is a set of independent measurements.
    counts : ndarray shape = (Number of measurements, NDetectors**NQubits )
        Each row in the matrix is a set of independent measurements.
    crosstalk : ndarray shape = ( Number of measurements , 2**NQubits,2**NQubits ) (optional)
        The crosstalk matrix to compensate for inefficentcies in your beam splitter.
    efficiency : 1darray with length = NQubits*NDetectors (optional)
        The relative efficienies between your detector pairs.
    time :  1darray with length = Number of measurements (optional)
        The total duration of each measurment.
    singles : ndarray shape = ( Number of measurements , 2*NQubits ) (optional)
        The singles counts on each detector.
    window : 1darray with length = NQubits*NDetectors (optional)
        The coincidence window duration for each detector pair.
    error : int (optional)
        The number of monte carlo states used to estimate the properties of the state.
    intensities : 1darray with length = Number of measurements (optional)
        The relative intensity of each measurement. Used for drift correction.
    method : ['MLE','HMLE','BME','LINER'] (optional)
        Which method to use to run tomography. Default is MLE

    Returns
    -------
    tomo_input : ndarray
        The input data for the current tomography.
    intensities : 1darray with length = number of measurements
        Relative pump power (arb. units) during measurement; used for drift correction. Default will be an array of ones
    """
    def buildTomoInput(self, measurements, counts, crosstalk, efficiency,time,singles,window, error):
        ################
        # measurements #
        ################
        # Get the number of qubits based on the measurement matrix
        if (len(measurements.shape)!=2 or measurements.shape[1]%2 != 0):
            raise ValueError("Invalid measurements matrix")
        else:
            self.conf['NQubits'] = int(measurements.shape[1]/2)
        # TODO: make sure measurements span the entire space

        ##########
        # counts #
        ##########
        # Check if counts has right dimensions
        if (counts.shape[0] != measurements.shape[0]):
            raise ValueError("Number of Counts does not match the number of measurements")
        # Determine if 2det/qubit from counts matrix
        try:
            if(counts.shape[1] == 1):
                self.conf['NDetectors'] = 1
            elif(counts.shape[1] == self.conf['NQubits']*2):
                self.conf['NDetectors'] = 2
            else:
                raise ValueError("The second axis of counts does not have the right dimension. Should be 1 or 2*NQubits for 2det")
        except:
            self.conf['NDetectors'] = 1

        ##############
        # efficiency #
        ##############
        if((not isinstance(efficiency, int)) and (len(efficiency.shape) !=1 or efficiency.shape[0] != self.conf['NQubits']*2)):
            raise ValueError("Invalid efficiency array. Length should be NQubits*2")
        else:
            self.conf['Efficiency'] = efficiency

        ##########
        # window #
        ##########
        if ((not isinstance(window, int)) and (len(window.shape) != 1 or window.shape[0] != self.getNumCoinc())):
            raise ValueError("Invalid window array. Length should be NQubits*NDetectors")
        else:
            self.conf['Window'] = window

        ################
        # time/singles #
        ################
        if ((not isinstance(time, int)) or (not isinstance(singles, int)) or (not isinstance(window, int))):
            self.conf['DoAccidentalCorrection'] = 1
            # Check if time has right length
            if (isinstance(time, int)):
                time = np.ones(measurements.shape[0])
            elif not (len(time.shape) == 1 and time.shape[0] == measurements.shape[0]):
                ValueError("Invalid time array")
            # Check if singles has right dimensions
            if (isinstance(singles, int)):
                if self.conf['NDetectors'] == 1:
                    singles = np.zeros((measurements.shape[0], self.conf['NQubits']))
                elif self.conf['NDetectors'] == 2:
                    singles = np.zeros((measurements.shape[0], 2 * self.conf["NQubits"]))
            else:
                if self.conf['NDetectors'] == 1 and self.conf['NQubits'] == 1:
                    if singles.shape != (measurements.shape[0],) and singles.shape != (measurements.shape[0], 1):
                        raise ValueError("Invalid singles matrix")
                    # if the singles vector has the form (x,), this changes it to (x,1)
                    if np.ndim(singles) == 1:
                        singles = np.atleast_2d(singles).transpose()
                elif self.conf['NDetectors'] == 1 and self.conf['NQubits'] != 1:
                    if singles.shape != (measurements.shape[0], self.conf['NQubits']):
                        raise ValueError("Invalid singles matrix")
                elif self.conf['NDetectors'] == 2:
                    if singles.shape != (measurements.shape[0], 2 * self.conf['NQubits']):
                        raise ValueError("Invalid Singles matrix")
            # Check if window has right length
            if (isinstance(window, int)):
                self.conf['Window'] = window
            elif (len(window.shape) == 1 and window.shape[0] == self.getNumCoinc()):
                self.conf['Window'] = window
            else:
                raise ValueError("Invalid window array")
        else:
            time = np.ones(measurements.shape[0])
            singles = np.zeros((measurements.shape[0],self.conf["NQubits"]))
            self.conf['Window'] = window


        #############
        # crosstalk #
        #############
        if (isinstance(crosstalk, int)):
            self.conf['Crosstalk'] = np.identity(2 ** self.conf["NQubits"])
        elif (len(crosstalk.shape) == 2 and crosstalk.shape[0] == crosstalk.shape[1] and crosstalk.shape[0] == 2 ** self.conf["NQubits"] ):
            self.conf['Crosstalk'] =  crosstalk
        else:
            raise ValueError("Invalid crosstalk matrix")

        #########
        # error #
        #########
        self.conf['DoErrorEstimation'] = error

        ##############
        # tomo_input #
        ##############
        # Here we build the tomo_input matrix and then return this to be fed to StateTomography_Matrix
        n_qubit = self.conf['NQubits']
        if (self.conf['NDetectors'] == 1):
            tomo_input = np.zeros((measurements.shape[0], 3 * n_qubit + 2), dtype=complex)
            # times
            tomo_input[:, 0] = time
            # singles
            tomo_input[:, 1: n_qubit + 1] = singles
            # counts
            tomo_input[:, n_qubit + 1] = counts
            # measurements
            tomo_input[:, n_qubit + 2: 3 * n_qubit + 2] = measurements
        else:
            tomo_input = np.zeros((measurements.shape[0], 2 ** n_qubit + 4 * n_qubit + 1), dtype=complex)
            # times
            tomo_input[:, 0] = time
            # singles
            tomo_input[:, 1: 2 * n_qubit + 1] = singles
            # counts
            tomo_input[:, 2 * n_qubit + 1: 2 ** n_qubit + 2 * n_qubit + 1] = counts
            # measurements
            tomo_input[:, 2 ** n_qubit + 2 * n_qubit + 1: 2 ** n_qubit + 4 * n_qubit + 1] = measurements


        return tomo_input

    """
    checkForInvalidSettings()
    desc : Checks the current settings and throws errors if any are invalid.
    """
    def checkForInvalidSettings(self):
        # Check Accidentals and window
        if self.conf['nqubits'] == 1:
            if self.conf['DoAccidentalCorrection']:
                raise ValueError('Invalid Conf settings. Accidental Correction can not be done for single qubit tomography')
        if self.conf['DoAccidentalCorrection']:
            try:
                win = np.array(self.conf['window'],dtype=float)
                if len(win.shape) != 1 or len(win) != self.getNumCoinc():
                    raise
                self.conf['window'] = win
            except:
                raise ValueError('Invalid Conf settings. Window should have length ' +str(self.getNumCoinc()) + " with the given settings.")
        # Efficicieny and Ndetectors
        if not self.conf['NDetectors'] in [1,2]:
            raise ValueError('Invalid Conf settings. NDetectors can be either 1 or 2, corresponding to the number of detectors per qubit.')
        elif self.conf['NDetectors'] == 2:
            try:
                eff = np.array(self.conf['Efficiency'],dtype=float)
                if isinstance(eff,int):
                    eff = np.ones(self.getNumCoinc())
                if len(eff.shape) != 1 or len(eff) != self.getNumCoinc():
                    raise
            except:
                raise ValueError('Invalid Conf settings. Efficiency should have length ' +str(self.getNumCoinc()) + " with the given settings.")
        elif self.conf['NDetectors'] == 1:
            eff = np.ones(1)
        self.conf['Efficiency'] = eff
        # Crosstalk
        try:
            correctSize = int(np.floor(2**self.getNumQubits()+.01))
            c = self.conf['crosstalk']
            if isinstance(c,int):
                c = np.eye(correctSize)
            else:
                c = np.array(c)
            # make sure it has the right shape
            if len(c.shape) != 2 or c.shape[0] != c.shape[1]:
                raise ValueError('Invalid Conf settings. Crosstalk should be an array with shape (' +
                                 str(correctSize)+","+str(correctSize)+ ") with the given settings.")
            # make sure it has the right size
            if c.shape[0] != correctSize:
                if np.any(c-np.eye(c.shape[0])>10**-6):
                    # Throw error if not the right size
                    raise ValueError('Invalid Conf settings. Crosstalk should be an array with shape (' +
                                     str(correctSize) + "," + str(correctSize) + ") with the given settings.")
                else:
                    # If the given cross talk is an identity matrix just fix the size.
                    c = np.eye(correctSize)
            self.conf['crosstalk'] = c
        except:
            raise ValueError('Invalid Conf settings. Crosstalk should be an array with shape (' +
                             str(correctSize) + "," + str(correctSize) + ") with the given settings.")

    # # # # # # # # # #
    '''Get Functions'''
    # # # # # # # # # #

    """
    getCoincidences()
    Desc: Returns an array of counts for all the measurements.
    """
    def getCoincidences(self):
        if (self.conf['NDetectors'] == 2):
            return np.real(self.last_input[:, np.arange(2*self.conf['NQubits']+1, 2**self.conf['NQubits']+2*self.conf['NQubits']+1)])
        else:
            return np.real(self.last_input[:, self.conf['NQubits']+1])

    """
    getSingles()
    Desc: Returns an array of singles for all the measurements.
    """
    def getSingles(self):
        if (self.conf['NDetectors'] == 2):
            return np.real(self.last_input[:, np.arange(1, 2*self.conf['NQubits']+1)])
        else:
            return np.real(self.last_input[:, np.arange(1, self.conf['NQubits']+1)])

    """
    getTimes()
    Desc: Returns an array of times for all the measurements.
    """
    def getTimes(self):
        return self.last_input[:, 0]

    """
    getMeasurements()
    Desc: Returns an array of measurements in pure state form for all the measurements.
    """
    def getMeasurements(self):
        if (self.conf['NDetectors'] == 2):
            return self.last_input[:, np.arange(2**self.conf['NQubits']+2*self.conf['NQubits']+1, 2**self.conf['NQubits']+4*self.conf['NQubits']+1)]
        else:
            return self.last_input[:, np.arange(self.conf['NQubits']+2, 3*self.conf['NQubits']+2)]

    """
    getNumQubits()
    Desc: returns the number of qubits for the current configurations.
    """
    def getNumQubits(self):
        return self.conf['NQubits']

    """
    getNumCoinc()
    Desc: Returns the number of coincidences per measurement for the current configurations.
    """
    def getNumCoinc(self):
        if (self.conf['NDetectors'] == 2):
            return 2 ** self.conf['NQubits']
        else:
            return 1

    """
    getNumSingles()
    Desc: Returns the number of singles per measurement for the current configurations.
    """
    def getNumSingles(self):
        if (self.conf['NDetectors'] == 2):
            return 2 * self.conf['NQubits']
        else:
            return self.conf['NQubits']

    """
    getNumDetPerQubit()
    Desc: returns the number of detectors per qubit for the current configurations.
    """
    def getNumDetPerQubit(self):
        return self.conf['NDetectors']


    """
    getStandardBasis(numBit,numDet = -1)
    Desc: Returns an array of standard measurements in separated pure state form for the given number of qubits. It is the same format used in the tomo_input matrix.

    Parameters
    ----------
    numBits : int
        number of qubits you want for each measurement. Default will use the number of qubits in the current configurations.
    numDet : 1 or 2
        Number of detectors for each measurement Default will use the number of qubits in the current configurations.   
    """
    def getStandardBasis(self, numBits = -1,numDet = -1):
        # check if numBits is an int
        if not (isinstance(numBits, int)):
            raise ValueError("numBits must be an integer")
        if(numBits < 1):
            numBits = self.getNumQubits()

        # check if numDet is not 1 or 2
        if not (numDet in [-1, 1, 2]):
            raise ValueError("numDet must be an 1 or 2")
        if (numDet < 1):
            numDet = self.getNumDetPerQubit()

        # Define start meas basis
        if(numDet == 1):
            basis = np.array([[1, 0],
                              [0, 1],
                              [(2 ** (-1 / 2)), (2 ** (-1 / 2))],
                              [(2 ** (-1 / 2)), -(2 ** (-1 / 2))],
                              [(2 ** (-1 / 2)), (2 ** (-1 / 2)) * 1j],
                              [(2 ** (-1 / 2)), -(2 ** (-1 / 2)) * 1j]],dtype=complex)
        else:
            basis = np.array([[1, 0],
                              [(2 ** (-1 / 2)), (2 ** (-1 / 2))],
                              [(2 ** (-1 / 2)), (2 ** (-1 / 2)) * 1j]], dtype=complex)
        numBasis = basis.shape[0]
        m = np.zeros((numBasis ** numBits, 2 * numBits), dtype = complex)
        for i in range(m.shape[0]):
            for j in range(0, m.shape[1], 2):
                bitNumber = np.floor((m.shape[1] - j - 1) / 2)
                index = int(((i) / numBasis ** (bitNumber)) % numBasis)

                m[i, j] = basis[index][0]
                m[i, j + 1] = basis[index][1]
        return m

    """
    getTomoInputTemplate()
    Desc: returns a standard template for tomo_input for the given number of qubits.

    Parameters
    ----------
    numBits : int
        number of qubits you want for each measurement. Default will use the number of qubits in the current configurations.
    numDet : 1 or 2
        Number of detectors for each measurement Default will use the number of qubits in the current configurations.   

    Returns
    -------
    Tomoinput : ndarray
        The input data for the current tomography.Example can be seen at top of page.
    """
    def getTomoInputTemplate(self, numBits = -1,numDet = -1):
        # check if numBits is an int
        if not (isinstance(numBits, int)):
            raise ValueError("numBits must be an integer")
        if (numBits < 1):
            numBits = self.getNumQubits()

        # check if numDet is not 1 or 2
        if not(numDet in [-1,1,2]):
            raise ValueError("numDet must be an 1 or 2")
        if (numDet < 1):
            numDet = self.getNumDetPerQubit()

        measurements = self.getStandardBasis(numBits,numDet)

        if(numDet == 1):
            # For n detectors:
            Tomoinput = np.zeros((6**numBits, 3*numBits+2), dtype = complex)

            # input[:, n_qubit+1]: coincidences
            # coincidences left zeros

            # input[:, np.arange(n_qubit+2, 3*n_qubit+2)]: measurements
            Tomoinput[:, np.arange(numBits + 2, 3 * numBits + 2)] = measurements

        else:
            # For 2n detectors:
            Tomoinput = np.zeros((3**numBits, 2**numBits+4*numBits+1), dtype = complex)

            # input[:, np.arange(2*n_qubit+1, 2**n_qubit+2*n_qubit+1)]: coincidences
            # coincidences left zeros

            # input[:, np.arange(2**n_qubit+2*n_qubit+1, 2**n_qubit+4*n_qubit+1)]: measurements
            Tomoinput[:, np.arange(2**numBits+2*numBits+1, 2**numBits+4*numBits+1)] = measurements

        # tomo_input[:, 0]: times
        Tomoinput[:, 0] = np.ones_like(Tomoinput[:, 0])

        # tomo_input[:, np.arange(1, 2 * n_qubit + 1)]: singles
        # singles left zero
        return Tomoinput

    # # # # # # # # # # #
    '''ERROR FUNCTIONS'''
    # # # # # # # # # # #

    """
    getProperties(bounds)
    Desc: Returns all the properties of the given density matrix.

    Parameters
    ----------
    bounds : int (optional)
        The number of monte carlo runs you want to perform to get a better estimate of each property. Default will use whatever is set in the conf settings.

    Returns
    -------
    vals : ndarray with shape = (length of self.err_functions, 2)
        The first col is the name of the property.
        The second col is the value of the property.
        The third col is the error bound on the property.
    """
    def getProperties(self,bounds = -1):
        # if bounds not set use the conf settings
        if bounds == -1:
            bounds = self.conf['DoErrorEstimation']

        # generate states if needed
        if bounds > len(self.mont_carlo_states)-1:
            self.tomography_states_generator(bounds-len(self.mont_carlo_states)+1)
        states = np.array(self.mont_carlo_states,dtype="O")[:bounds + 1, :]
        if states.shape[0] == 1:
            vals1 = np.array([["intensity",np.mean(states[:,1]),"NA"],
                              ["fval", np.mean(states[:,2]),"NA"]],dtype="O")
        else:
            vals1 = np.array([["intensity", np.mean(states[:, 1]), np.std(states[:, 1],ddof=1)],
                              ["fval", np.mean(states[:, 2]), np.std(states[:, 2],ddof=1)]],dtype="O")
        vals2 = getProperties_helper_bounds(self.err_functions, states[:,0])

        vals = np.concatenate((vals1, vals2))
        return vals


    """
    tomography_states_generator(n)
    Desc: Uses monte carlo simulation to create random states similar to the estimated state.
          The states are also saved under self.mont_carlo_states

    Parameters
    ----------
    n : int
        Number of approximate states you want to create.
        If no value is given it will use the conf settings.

    Returns
    -------
    rhop : ndarray with shape = (n, 2^numQubits, 2^numQubits)
        The approximate density matrices.
    intenp : 1darray with length = n
        The intensity of each approximate density matrices.
    fvalp : 1darray with length = n
        The fval of the regression associated with each approximate density matrices.
    """
    def tomography_states_generator(self, n):
        # Save the last data so we can restore it later
        last_input = self.last_input.copy()

        ndet = self.conf['NDetectors']
        acc = self.conf['DoAccidentalCorrection']
        time = self.getTimes()
        meas = self.getMeasurements()
        singles = self.getSingles()
        counts = self.getCoincidences()

        # Re-sample counts and singles
        test_counts = np.random.poisson(counts)
        test_singles = np.random.poisson(singles)
        for j in range(n):
            if len(test_counts.shape) == 1:
                test_counts = np.array([test_counts]).T
            test_data = np.concatenate((np.array([time]).T, test_singles, test_counts, meas), axis = 1)
            [rhop, intenp, fvalp] = self.StateTomography_Matrix(test_data, self.intensities,method=self.conf['method'],_saveState=False)
            self.mont_carlo_states.append([rhop, intenp, fvalp])
        # Restore the last tomo_input matrix to the original one
        self.last_input = last_input
        return self.mont_carlo_states

    """
    getBellSettings(bounds)
    Desc: Returns the optimal measurment settings for the CHSH bell inequality. In Progress, have not checked.
    """
    def getBellSettings(self, partsize_init = 9, partsize = 5, t = 3, bounds = -1):
        if self.conf['nqubits'] != 2:
            raise ValueError('Invalid Conf settings. Bell state can only be found for 2 qubit tomography')
        rho = self.last_rho
        # if bounds not set use the conf settings
        if bounds == -1:
            bounds = self.conf['DoErrorEstimation']

        # generate states if needed
        if bounds > len(self.mont_carlo_states) - 1:
            self.tomography_states_generator(bounds - len(self.mont_carlo_states) + 1)
        states = np.array(self.mont_carlo_states,dtype="O")[:bounds + 1, :]

        if (bounds > 0):
            return getBellSettings_helper_bounds(states[:,0], rho, partsize_init, partsize, t, bounds)
        else:
            return getBellSettings_helper(rho, partsize_init, partsize, t)

    """
    printLastOutput(bounds)
    Desc: Prints the properties of the last tomography to the console. Properties are defined in tomography conf settings. The calculated properties are determined by self.err_functions.

    Parameters
    ----------
    bounds : int (optional)
        The number of monte carlo runs you want to perform to get a better estimate of each property. Default will use whatever is set in the conf settings.
    """
    def printLastOutput(self, bounds = -1):
        p = np.array(self.last_rho.copy(), dtype = "O")
        print("State: ")
        mx = 0
        for i in range(p.shape[0]):
            for j in range(p.shape[1]):
                p[i, j] = floatToString(p[i, j]).replace(" ", "") + "  "
                if(len(p[i, j])>mx):
                    mx = len(p[i, j])

        for i in range(p.shape[0]):
            for j in range(p.shape[1]):
                print(p[i, j] + " "*(mx-len(p[i, j])), end = "")
            print("")

        # print(p)
        properties = self.getProperties(bounds)
        for prop in properties:
            if(len(prop) >=3) and prop[2] != 'NA':
                print(prop[0] + " : " + floatToString(prop[1]) + " +/- " + floatToString(prop[2]))
            else:
                print(prop[0] + " : " + floatToString(prop[1]))

    """
    exportToEval(filePath)
    Desc: Exports the input data to the specified file path. You can rerun tomography on this file using importEval() command.

    Parameters
    ----------
    filePath : str (optional)
        The file name you want the data saved to.

    See Also
     ------ 
        importEval
    """
    def exportToEval(self, filePath="pythonEval.txt"):
        TORREPLACE = ""
        # Conf settings
        for k in self.conf.keys():
            if k == "method":
                TORREPLACE += "conf['" + str(k) + "'] = '" + str(self.conf[k]) + "'\n"
            elif (isinstance(self.conf[k], np.ndarray)):
                A = self.conf[k]
                TORREPLACE += "conf['" + str(k) + "'] = np.array([\n"
                for i in range(A.shape[0]):
                    if len(A.shape) == 2:
                        TORREPLACE += "["
                        for j in range(A.shape[1]):
                            TORREPLACE += str(A[i, j]) + ","
                        TORREPLACE = TORREPLACE[:-1] + "],\n"
                    else:
                        TORREPLACE += str(A[i]) + ","
                if len(A.shape) == 2:
                    TORREPLACE = TORREPLACE[:-2] + "])\n"
                else:
                    TORREPLACE = TORREPLACE[:-1] + "])\n"
            else:
                TORREPLACE += "conf['" + str(k) + "'] = " + str(self.conf[k]) + "\n"

        # Tomoinput
        A = self.last_input
        TORREPLACE += "tomo_input = np.array([\n"
        for i in range(A.shape[0]):
            TORREPLACE += "["
            for j in range(A.shape[1]):
                TORREPLACE += str(A[i, j]) + ","
            TORREPLACE = TORREPLACE[:-1] + "],\n"

        TORREPLACE = TORREPLACE[:-2] + "])\n"

        # intensity
        A = self.intensities
        TORREPLACE += "intensity = np.array(["
        for i in range(A.shape[0]):
            TORREPLACE += str(A[i]) + ","
        TORREPLACE = TORREPLACE[:-1] + "])"

        # print contents to file
        with open(filePath, 'w') as f:
            f.write(TORREPLACE)

    """
    exportToConf(filePath)
    Desc: Exports the conf data to the specified file path. You can rerun tomography on this file using importConf() and importData() command.

    Parameters
    ----------
    filePath : str (optional)
        The file name you want the data saved to.

    See Also
     ------ 
        exportToData;importConf;importData
    """
    def exportToConf(self, filePath="pythonConf.txt"):
        TORREPLACE = ""
        # Conf settings
        for k in self.conf.keys():
            if k == "method":
                TORREPLACE += "conf['" + str(k) + "'] = '" + str(self.conf[k]) + "'\n"
            elif (isinstance(self.conf[k], np.ndarray)):
                A = self.conf[k]
                TORREPLACE += "conf['" + str(k) + "'] = np.array([\n"
                for i in range(A.shape[0]):
                    if len(A.shape) == 2:
                        TORREPLACE += "["
                        for j in range(A.shape[1]):
                            TORREPLACE += str(A[i, j]) + ","
                        TORREPLACE = TORREPLACE[:-1] + "],\n"
                    else:
                        TORREPLACE += str(A[i]) + ","
                if len(A.shape) == 2:
                    TORREPLACE = TORREPLACE[:-2] + "])\n"
                else:
                    TORREPLACE = TORREPLACE[:-1] + "])\n"
            else:
                TORREPLACE += "conf['" + str(k) + "'] = " + str(self.conf[k]) + "\n"

            # print contents to file
            with open(filePath, 'w') as f:
                f.write(TORREPLACE)

    """
     exportToData(filePath)
     Desc: Exports the tomo_input matrix  data to the specified file path. You can rerun tomography on this file using importConf() and importData() command.

     Parameters
     ----------
     filePath : str (optional)
         The file name you want the data saved to.

     See Also
      ------ 
         exportToConf;importConf;importData
     """
    def exportToData(self, filePath="pythonData.txt"):
        TORREPLACE = ""

        # Tomoinput
        A = self.last_input
        TORREPLACE += "tomo_input = np.array([\n"
        for i in range(A.shape[0]):
            TORREPLACE += "["
            for j in range(A.shape[1]):
                TORREPLACE += str(A[i, j]) + ","
            TORREPLACE = TORREPLACE[:-1] + "],\n"

        TORREPLACE = TORREPLACE[:-2] + "])\n"

        # intensity
        A = self.intensities
        TORREPLACE += "intensity = np.array(["
        for i in range(A.shape[0]):
            TORREPLACE += str(A[i]) + ","
        TORREPLACE = TORREPLACE[:-1] + "])"

        # print contents to file
        with open(filePath, 'w') as f:
            f.write(TORREPLACE)

        # print contents to file
        with open(filePath, 'w') as f:
            f.write(TORREPLACE)

    '''
    exportToConf_web(filePath)
    Desc: Exports the conf data to the specified file path. 
    You can run this tomography on our tomography interface at http://tomography.web.engr.illinois.edu/TomographyDemo.php
    THIS IS A WORK IN PROGRESS.

    Parameters
    ----------
    filePath : str (optional)
        The file name you want the data saved to.

    See Also
     ------ 
        exportToData_web
    '''
    def exportToConf_web(self, filePath="Config_web.txt"):
        # Conf settings
        TORREPLACE = "conf.NQubits="+str(self.conf['NQubits'])+";\n" \
                     "conf.NDetectors="+str(self.conf['NQubits'])+";\n" \
                     "conf.Crosstalk=["
        A = self.conf["crosstalk"]
        for i in range(len(A)):
            TORREPLACE += "["
            for j in range(len(A[i])):
                TORREPLACE += str(A[i][j]).replace(" ", "") + ","
            TORREPLACE = TORREPLACE[:-1] + "],"
        TORREPLACE = TORREPLACE[:-1] + "];\n"

        TORREPLACE +="conf.UseDerivative=0;\n"
        if self.conf['Bellstate']:
            TORREPLACE +="conf.Bellstate=1;\n"
        else:
            TORREPLACE += "conf.Bellstate=0;\n"
        A = self.conf["Window"]
        if not (isinstance(A,np.ndarray) or isinstance(A,list)):
            A = [A]
        TORREPLACE += "conf.Window=["
        for i in range(len(A)):
            TORREPLACE += str(A[i]).replace(" ", "") + ","
        TORREPLACE += TORREPLACE[:-1] + "];\n"
        A = self.conf["Efficiency"]
        if not (isinstance(A,np.ndarray) or isinstance(A,list)):
            A = [A]
        TORREPLACE += "conf.Efficiency=["
        for i in range(len(A)):
            TORREPLACE += str(A[i]).replace(" ", "") + ","
        TORREPLACE += TORREPLACE[:-1] + "];\n"
        # print contents to file
        with open(filePath, 'w') as f:
            f.write(TORREPLACE)

    '''
    exportToData_web(filePath)
    Desc: Exports the tomo_input matrix data to the specified file path. 
    You can run this tomography on our tomography interface at http://tomography.web.engr.illinois.edu/TomographyDemo.php
    THIS IS A WORK IN PROGRESS.
    
    Parameters
    ----------
    filePath : str (optional)
        The file name you want the data saved to.

    See Also
     ------ 
        exportToConf_web
    '''
    # def exportToData_web(self, filePath="pythonData.txt"):
    #     TORREPLACE = ""
    #
    #     # Tomoinput
    #     A = self.last_input.astype("O")
    #     n_qubit = self.conf['NQubits']
    #     if self.conf['NDetectors'] == 2:
    #         # tomo_input[ :, 1 : 2 * n_qubit + 1 ]: singles
    #         A[:, 1: 2 * n_qubit + 1] = self.last_input[:, 1: 2 * n_qubit + 1].real.astype(int)
    #         # tomo_input[ :, 2 * n_qubit + 1 : 2 ** n_qubit + 2 * n_qubit + 1 ]: coincidences
    #         A[:, 2 * n_qubit + 1: 2 ** n_qubit + 2 * n_qubit + 1] = self.last_input[:,
    #                                                                 2 * n_qubit + 1: 2 ** n_qubit + 2 * n_qubit + 1].real.astype(
    #             int)
    #     else:
    #         # tomo_input[:, 1: n_qubit + 1]: singles
    #         A[:, 1: n_qubit + 1] = self.last_input[:, 1: n_qubit + 1].real.astype(int)
    #         # tomo_input[:, n_qubit + 1]: coincidences
    #         A[:, n_qubit + 1] = self.last_input[:, n_qubit + 1].real.astype(int)
    #
    #     TORREPLACE += "tomo_input=["
    #     for i in range(A.shape[0]):
    #         TORREPLACE += "["
    #         for j in range(A.shape[1]):
    #             if isinstance(A[i, j],int):
    #                 TORREPLACE += str(A[i, j]) + ","
    #             else:
    #                 TORREPLACE += floatToString(A[i, j]) + ","
    #         TORREPLACE = TORREPLACE[:-1] + "],"
    #
    #     TORREPLACE = TORREPLACE[:-1] + "];\n"
    #
    #     # intensity
    #     A = self.intensities
    #     TORREPLACE += "intensity=["
    #     for i in range(A.shape[0]):
    #         TORREPLACE += str(A[i]) + ","
    #     TORREPLACE = TORREPLACE[:-1] + "];"
    #
    #     # print contents to file
    #     with open(filePath, 'w') as f:
    #         f.write(TORREPLACE)



from __future__ import print_function
import numpy as np
from .TomoFunctions import *

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""


# These are various helper functions used in other main functions



###############
# LIKELYHOODS #
###############

"""
maxlike_fitness(t, coincidences, accidentals, m, prediction)
Desc: Calculates the diffrence between the current predicted state data and the actual data.

Parameters
----------
t : ndarray
    T values of the current predicted state.
coincidences : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The counts of the tomography.
accidentals : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The singles values of the tomography. Used for accidental correction.
m : ndarray with shape = (2^numQubits, 2^numQubits, number of measurements)
    The measurements of the tomography in density matrix form.
prediction : ndarray
    Predicted counts from the predicted state.
overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
    The relative weights of each measurment. Used for drift correction.
    
Returns
-------
val : float
    value of the optimization function.
"""
def maxlike_fitness(t, coincidences, accidentals, measurements, overall_norms):
    rhog = t_to_density(t,normalize=False)
    prediction = np.zeros_like(coincidences)
    for j in range(len(prediction)):
        prediction[j] = overall_norms[j] * np.real(np.trace(np.dot(measurements[j, :, :], rhog))) + accidentals[j]
        prediction[j] = np.max([prediction[j], 0.01])
    log_like = (prediction - coincidences) / np.sqrt(prediction)
    return np.real(log_like)
"""
maxlike_fitness_hedged(t, coincidences, accidentals, measurements, prediction, bet)
Desc: Calculates the diffrence between the current predicted state data and the actual data using hedged maximum likelihood.

Parameters
----------
t : ndarray
    T values of the current predicted state.
coincidences : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The counts of the tomography.
accidentals : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
    The singles values of the tomography. Used for accidental correction.
measurements : ndarray with shape = (2^numQubits, 2^numQubits, number of measurements)
    The measurements of the tomography in density matrix form .
prediction : ndarray
    Predicted counts from the predicted state.
bet : float 0 to .5
    The beta value used.
overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
    The relative weights of each measurment. Used for drift correction.

Returns
-------
val : float
    value of the optimization function.
"""
def maxlike_fitness_hedged(t, coincidences, accidentals, measurements, bet,overall_norms):
    prediction = np.zeros_like(coincidences)
    rhog = t_to_density(t,normalize=False)
    for j in range(len(prediction)):
        prediction[j] = overall_norms[j] * np.real(np.trace(np.dot(measurements[j,:,:], rhog))) + accidentals[j]
        prediction[j] = np.max([prediction[j], 0.01])
    hedge = np.repeat(np.real((bet * np.log(np.linalg.det(rhog))) / len(prediction)), len(prediction))
    val = np.sqrt(np.real((((prediction - coincidences) ** 2) / (2 * prediction)) - hedge) + 1000)
    return np.real(val)

# This are some old functions that have since been reworked.
#
# """
# maxlike_fitness_old(t, coincidences, accidentals, m, prediction)
# Desc: Calculates the diffrence between the current predicted state data and the actual data
#
# Parameters
# ----------
# t : ndarray
#     T values of the current predicted state.
# coincidences : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
#     The counts of the tomography.
# accidentals : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
#     The singles values of the tomography. Used for accidental correction.
# m : ndarray with shape = (2^numQubits, 2^numQubits, number of measurements)
#     The measurements of the tomography in density matrix form.
# prediction : ndarray
#     Predicted counts from the predicted state.
# overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
#     The relative weights of each measurment. Used for drift correction.
#
# Returns
# -------
# val : float
#     value of the optimization function.
# """
# def maxlike_fitness_old(t, coincidences, accidentals, m, prediction,overall_norms=-1):
#     rhog = t_to_density(t,normalize=False)
#     for j in range(len(prediction)):
#         prediction[j] = np.float64(np.real(overall_norms[j] * np.real(np.trace(np.dot(m[:, :, j], rhog))) + accidentals[j]))
#         prediction[j] = np.max([prediction[j], 0.01])
#     val = (prediction - coincidences) / np.sqrt(prediction)
#     val = np.float64(np.real(val))
#     return val
#
#
# """
# maxlike_fitness_hedged_old(t, coincidences, accidentals, m, prediction, bet)
# Desc: Calculates the diffrence between the current predicted state data and the actual data using hedged maximum likelihood.
#
# Parameters
# ----------
# t : ndarray
#     T values of the current predicted state.
# coincidences : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
#     The counts of the tomography.
# accidentals : ndarray with length = number of measurements or shape = (number of measurements, 2^numQubits) for 2 det/qubit
#     The singles values of the tomography. Used for accidental correction.
# m : ndarray with shape = (2^numQubits, 2^numQubits, number of measurements)
#     The measurements of the tomography in density matrix form .
# prediction : ndarray
#     Predicted counts from the predicted state.
# bet : float 0 to .5
#     The beta value used.
# overall_norms : 1darray with length = number of measurements or length = number of measurements * 2^numQubits for 2 det/qubit
#     The relative weights of each measurment. Used for drift correction.
#
# Returns
# -------
# val : float
#     value of the optimization function.
# """
# def maxlike_fitness_hedged_old(t, coincidences, accidentals, m, prediction, bet,overall_norms=-1):
#     # If overall_norms not given then assume uniform
#     if not isinstance(overall_norms,np.ndarray):
#         overall_norms = np.ones(coincidences.shape[0])
#     elif not (len(overall_norms.shape) == 1 and overall_norms.shape[0] == coincidences.shape[0]):
#         raise ValueError("Invalid intensities array")
#
#     rhog = t_to_density(t,normalize=False)
#     for j in range(len(prediction)):
#         prediction[j] = overall_norms[j] * np.real(np.trace(np.dot(m[:, :, j], rhog))) + accidentals[j]
#         prediction[j] = np.max([prediction[j], 0.01])
#
#     hedge = np.repeat(np.real((bet * np.log(np.linalg.det(np.mat(rhog)))) / len(prediction)), len(prediction))
#     val = np.sqrt(np.real((((prediction - coincidences) ** 2) / (2 * prediction)) - hedge) + 1000)
#
#     val = np.float64(np.real(val))
#
#     return val

# helper function used in linear tomography
def independent_set(measurements):
    m = measurements[0, :].conj().transpose()
    # may have to switched order of m and measurements, may be wrong but has little effect
    matrix = rho2stokes(np.outer(m, measurements[0, :]))
    max_rank = matrix.shape[0]

    if (measurements.shape[0]) == max_rank:
        s = np.ones([measurements.shape[0], 1])
        return s

    s = np.zeros([measurements.shape[0], 1])
    s[0] = 1
    cur_rank = 1
    for j in np.arange(1, measurements.shape[0], 1):
    # may have to switched order of m and measurements, may be wrong but has little effect
        m = measurements[j, :].conj().transpose()
        sv = rho2stokes(np.outer(m, measurements[j, :]))
        if (np.linalg.matrix_rank(np.concatenate((matrix, sv), axis = 1), tol = 0.001)) > cur_rank:
            matrix = np.concatenate((matrix, sv), axis = 1)
            cur_rank += 1
            s[j] = 1
        else:
            s[j] = 0
        if cur_rank == max_rank:
            break
    return s

# This function converts a list of loglikelihoods to normalized likelihoods. Also returns the index of
# the min loglike
def normalizeLikelihoods(likelihoods):
    nIndex = np.argmin(likelihoods)
    nFactor = likelihoods[nIndex]
    scaled = likelihoods - nFactor
    likelihoods = np.exp(-1 * scaled)
    likelihoods = likelihoods/sum(likelihoods)
    return likelihoods, nIndex,scaled


# helper function that formats the projects into a single matrix
def b_matrix(projectors):
    dim_m = projectors.shape[1]
    dim_b = dim_m**2
    tmp = np.zeros([dim_b, dim_b])+0j
    for i in range(dim_b):
        for j in range(dim_b):
            tmp[i][j] = np.inner(projectors[i], np.inner(sigma_n(j, dim_m), projectors[i].conj().transpose()))
    b = tmp.transpose()
    return b


# helper function that formats the measurements into a single matrix
def m_matrix(mu, projectors, b_inv):
    dim_m = projectors.shape[1]
    dim_b = dim_m**2

    tmp = np.zeros([dim_m, dim_m])
    for j in range(dim_b):
        tmp = tmp + b_inv[mu][j]*sigma_n(j, dim_m)
    m = tmp

    return m

# # Helper function for looping through a multi indexed array
# def multiloop_index(j, lengths):
#     ind = np.zeros(len(lengths))
#     for k in range(len(lengths)-1):
#         sz = np.prod(lengths[np.arange(k+1, len(lengths))])
#         ind[k] = np.fix(j/sz)+1
#         j % = sz
#     ind[len(ind)-1] = j+1
#
#     return ind

def make_positive(rhog_in):
    d, v = np.linalg.eig(rhog_in)
    rhog = np.zeros(rhog_in.shape)
    for j in range(len(d)):
        rhog = rhog + np.abs(d[j])*np.outer(v[:, j], v[:, j].conj().transpose())
    rhog = (rhog + rhog.conj().transpose())/2.0
    return rhog

# helper function for re formatting matrices
def sigma_n(j, nn):
    if j < 0 or j > nn**2-1:
        print('sigma_N: j out of range for SU(N)')

    m = int(np.fix(j/nn))
    n = int(j % nn)
    tmp1 = np.zeros([nn, 1])
    tmp2 = np.zeros([nn, 1])
    tmp1[m] = 1
    tmp2[n] = 1

    if m < n:
        matrix = (np.outer(tmp1, tmp2.conj().transpose())+np.outer(tmp2, tmp1.conj().transpose()))*np.sqrt(nn/2.0)
    elif m > n:
        matrix = 1j*(np.outer(tmp1, tmp2.conj().transpose())-np.outer(tmp2, tmp1.conj().transpose()))*np.sqrt(nn/2.0)
    elif (m+1) < nn:
        z = np.zeros(nn)
        for i in range(m+1):
            z[i] = 1
        matrix = -(np.sqrt(nn/((m+1.0)**2+m+1.0)))*np.diag(z)
        matrix[m+1, m+1] = (m+1.0)*(np.sqrt(nn/((m+1.0)**2+m+1.0)))
    else:  # n = m = N
        matrix = np.identity(nn)

    return matrix

# Helper function for independent_set function
def rho2stokes(rhog):
    if rhog.ndim == 1:
        rhog = np.outer(rhog, rhog.conj().transpose())

    d = len(rhog)
    n = d**2

    ss = np.zeros([n, 1])+0j
    for j in range(n):
        ss[j] = np.trace(np.inner(rhog, sigma_n(j, d)))

    return ss

# Helper function for calculating the bell settings
def coinmat(a, b):
    k = np.array([np.cos(a)*np.cos(b), np.cos(a)*np.sin(b), np.sin(a)*np.cos(b), np.sin(a)*np.sin(b)])
    cmat = np.outer(k, k)

    return cmat

# Evaluates the given function with the given arguments
def fevel(funcname, *args):
    return eval(funcname)(*args)

# Helper function that returns the given properties of the given rho.
# This will not given you fval and intensity. So it is recommended to use getProperties in the TomoClass
def getProperties_helper(errf, rho0):
    return np.array([[errf, fevel(errf, rho0)]for errf in errf], dtype = 'O')


# Helper function that returns the given properties with bounds of the given rhos.
# This will not given you fvals and intensities. So it is recommended to use getProperties in the TomoClass
def getProperties_helper_bounds(errf, rhos):
    n = rhos.shape[0]
    n_fun = len(errf)
    data = np.zeros([n, n_fun], dtype = "O")
    for j in range(n):
        data[j] = getProperties_helper(errf, rhos[j])[:, 1]

    stds = np.zeros(n_fun, dtype = "O")
    means = np.zeros(n_fun, dtype = "O")

    for k in range(n_fun):
        if all(data[:, k] == 'NA'):
            stds[k] = 'NA'
            means[k] = 'NA'
        else:
            if n < 2:
                stds[k] = "NA"
                means[k] = np.mean(data[:, k])
            else:
                stds[k] = np.std(data[:, k], ddof = 1)
                means[k] = np.mean(data[:, k])

    return np.array([np.array(errf), means, stds], dtype = "O").transpose()

# used in getBellSettings
def bellsettings_range_init(rhog, partsize):
    sval = 0
    aval = 0
    apval = 0
    bval = 0
    bpval = 0

    for a in np.linspace(0, np.pi / 2, partsize):
        for ap in np.linspace(a, np.pi / 2, partsize):
            for b in np.linspace(0, np.pi / 2, partsize):
                for bp in np.linspace(b, np.pi / 2, partsize):
                    npp = np.real(np.trace(np.dot(coinmat(a, b), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(a + np.pi / 2, b + np.pi / 2), rhog)))
                    e_ab = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(ap, b), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(ap + np.pi / 2, b + np.pi / 2), rhog)))
                    e_apb = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(a, bp), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(a + np.pi / 2, bp + np.pi / 2), rhog)))
                    e_abp = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(ap, bp), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(ap + np.pi / 2, bp + np.pi / 2), rhog)))
                    e_apbp = 2 * (npp + nmm) - 1

                    s = e_ab + e_abp + e_apb + e_apbp - 2 * np.min([e_ab, e_abp, e_apb, e_apbp])

                    if s > sval:
                        sval = s
                        aval = a
                        apval = ap
                        bval = b
                        bpval = bp

    arange_s = [np.max([aval - ((np.pi / 2) / partsize), 0]), np.min([aval + ((np.pi / 2) / partsize), np.pi / 2])]
    aprange_s = [np.max([apval - ((np.pi / 2) / partsize), 0]),
                 np.min([apval + ((np.pi / 2) / partsize), np.pi / 2])]
    brange_s = [np.max([bval - ((np.pi / 2) / partsize), 0]), np.min([bval + ((np.pi / 2) / partsize), np.pi / 2])]
    bprange_s = [np.max([bpval - ((np.pi / 2) / partsize), 0]),
                 np.min([bpval + ((np.pi / 2) / partsize), np.pi / 2])]

    return [sval, arange_s, brange_s, aprange_s, bprange_s]

# used in getBellSettings
def bellsettings_range(rhog, partsize, arange, brange, aprange, bprange):

    sval = 0
    aval = 0
    apval = 0
    bval = 0
    bpval = 0

    for a in np.linspace(arange[0], arange[1], partsize):
        for ap in np.linspace(aprange[0], aprange[1], partsize):
            for b in np.linspace(brange[0], brange[1], partsize):
                for bp in np.linspace(bprange[0], bprange[1], partsize):
                    npp = np.real(np.trace(np.dot(coinmat(a, b), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(a + np.pi / 2, b + np.pi / 2), rhog)))
                    e_ab = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(ap, b), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(ap + np.pi / 2, b + np.pi / 2), rhog)))
                    e_apb = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(a, bp), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(a + np.pi / 2, bp + np.pi / 2), rhog)))
                    e_abp = 2 * (npp + nmm) - 1

                    npp = np.real(np.trace(np.dot(coinmat(ap, bp), rhog)))
                    nmm = np.real(np.trace(np.dot(coinmat(ap + np.pi / 2, bp + np.pi / 2), rhog)))
                    e_apbp = 2 * (npp + nmm) - 1

                    s = e_ab + e_abp + e_apb + e_apbp - 2 * np.min([e_ab, e_abp, e_apb, e_apbp])

                    if s > sval:
                        sval = s
                        aval = a
                        apval = ap
                        bval = b
                        bpval = bp

    arange_s = [np.max([aval - ((arange[1] - arange[0]) / partsize), 0]),
                np.min([aval + ((arange[1] - arange[0]) / partsize), np.pi / 2])]
    aprange_s = [np.max([apval - ((aprange[1] - aprange[0]) / partsize), 0]),
                 np.min([apval + ((aprange[1] - aprange[0]) / partsize), np.pi / 2])]
    brange_s = [np.max([bval - ((brange[1] - brange[0]) / partsize), 0]),
                np.min([bval + ((brange[1] - brange[0]) / partsize), np.pi / 2])]
    bprange_s = [np.max([bpval - ((bprange[1] - bprange[0]) / partsize), 0]),
                 np.min([bpval + ((bprange[1] - bprange[0]) / partsize), np.pi / 2])]

    return [sval, aval, apval, bval, bpval, arange_s, brange_s, aprange_s, bprange_s]

# Helper function that returns the optimal CHSH bell measurement settings given rho.
# It is recommended to use getBellSettings in the TomoClass
def getBellSettings_helper(rhog, partsize_init, partsize, t):

    [s, arange, brange, aprange, bprange] = bellsettings_range_init(rhog, partsize_init)
    a = 0
    b = 0
    ap = 0
    bp = 0

    for j in range(t):
        [s, a, ap, b, bp, arange, brange, aprange, bprange] = \
            bellsettings_range(rhog, partsize, arange, brange, aprange, bprange)

    return np.array([["s", s], ["a", a], ["a'", ap], ["b", b], ["b'", bp]], dtype = "O")


# Helper function that returns the optimal CHSH bell measurement settings with bounds of the given rhos.
# It is recommended to use getBellSettings in the TomoClass
def getBellSettings_helper_bounds(rhop, rho, partsize_init, partsize, t, n):
    belldata = np.zeros([n+1, 5])

    for j in range(n):
        belldata[j] = getBellSettings_helper(rhop[j], partsize_init, partsize, t)[:, 1]
    [bellNames, belldata[-1]] = getBellSettings_helper(rho, partsize_init, partsize, t).transpose()
    bmeans = np.zeros(5)
    berrors = np.zeros(5)

    for m in range(5):
        berrors[m] = np.std(belldata[:, m], ddof = n - 1)
        bmeans[m] = np.mean(belldata[:, m])

    return np.array([bellNames, berrors, bmeans], dtype = "O").transpose()

# Calculates the weighted covariance. Used in bayesian tomography
def weightedcov(samples,weights):
    mean = np.dot(weights,samples)
    diff = samples-mean
    covariance = np.array([np.outer(d, d) for d in diff])
    covariance = np.dot(covariance.T,weights)
    return [mean,covariance]


import sys
from .TomoClass import *
from .TomoDisplay import *
"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""



'''
    Comments Should be formatted like the following:
    (this comment is written in single quotes so that it is not picked up by the documentation generator).

    function(parameter 1, parameter 2)
    Desc: This is where the description is written.
	
    Parameters
    ----------      (10 hyphens, one for each letter)
    parameter 1 : type
    description
    parameter 2 : type
    description
                    (make sure the white spaces between are deleted all the way to the left side of the screen)
    Returns
    -------     (7 hyphens, one for each letter)
    returnValue 1 : type
    description
    returnValue 2 : type
    description
	2
    See Also
     ------   (space, 6 hyphens, another space)
    function1;function2;function2      (these should be the functionTitle without parentheses, as it appears on the Table of Contents
'''


# # # # # # # # # #  #
# # TO THE DEVELOPER #
# # # # # # # # # #  #

# Everything in this code is based on the Matlab files you can download on our tomography website, a more detailed
# comments and definations of variables could be found in those m files, The function names are as same as possible.
#
# 'UseDerivatives' is used automatically when doing 1 detctor per qubit, and is not used when doing 2 detectors per qubit.
# This cause the program to run faster and provide more accurate results in both cases



from __future__ import print_function
import scipy as sp
from numpy.core.defchararray import add
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D
from .TomoDisplayHelpers import *
from .TomoFunctions import removeGlobalPhase
from matplotlib.colors import LinearSegmentedColormap
import warnings

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""


"""
makeRhoImages(p, plt_given, customColor)
Desc: Creates matlab plots of the density matrix.

Parameters
----------
p : ndarray with shape = (n, 2^numQubits, 2^numQubits)
    The density matrix you want to create plots of.
plt_given : matplotlib.pyplot
    Input pyplot for which the figures will be saved on to.
customColor : boolean
    Specify if you want our custom colorMap. Default is true

See Also
 ------ 
saveRhoImages
"""
def makeRhoImages(p, plt_given, customColor = True):
    # Set up
    numQubits = int(np.log2(p.shape[0]))
    xpos = np.zeros_like(p.flatten(), dtype = float)
    ypos = np.zeros_like(p.flatten(), dtype = float)
    for i in range(0, 2**numQubits):
        xpos[i*2**numQubits:(1+i)*2**numQubits] = .5+i
    for i in range(0, 2**numQubits):
        ypos[i::2**numQubits] = .5+i
    zpos = np.zeros_like(p.flatten(), dtype = float)
    # width of cols
    dx = .9*np.ones_like(xpos)
    dy = .9*np.ones_like(ypos)
    # custom color map
    n_bin = 100
    if(customColor):
        from matplotlib.colors import LinearSegmentedColormap
        cmap_name = 'my_list'
        colors = [(1 / 255.0, 221 / 255.0, 137 / 255.0),
                  (32 / 255.0, 151 / 255.0, 138 / 255.0),
                  (53 / 255.0, 106 / 255.0, 138 / 255.0),
                  (86 / 255.0, 33 / 255.0, 139 / 255.0),
                  (131 / 255.0, 75 / 255.0, 114 / 255.0),
                  (173 / 255.0, 114 / 255.0, 90 / 255.0),
                  (253 / 255.0, 187 / 255.0, 45 / 255.0)]
        colorMap = LinearSegmentedColormap.from_list(cmap_name, colors, N = n_bin)
    else:
        colorMap = plt.cm.jet
    norm = mpl.colors.Normalize(vmin = -1, vmax = 1)

    tickBase = ["H", "V"]
    tick = [""]
    for x in range(numQubits):
        newTick = np.zeros(len(tick)*2, dtype = "O")
        for i in range(len(tick)):
            for j in range(len(tickBase)):
                newTick[len(tick)*i +j] = tick[i] + tickBase[j]
        tick = newTick
    xTicks = ["|"+x+">" for x in tick]
    yTicks = ["|"+x+">" for x in tick]


    # Real Graph
    fig = plt_given.figure()
    ax1 = fig.add_subplot(111, projection = '3d')
    dz = p.flatten().astype(float)
    img = ax1.bar3d(xpos, ypos, zpos, dx, dy, dz, color = colorMap((dz + 1) / 2), edgecolor = "black", alpha = .8)



    ax1.axes.set_xticklabels(xTicks)
    ax1.axes.set_yticklabels(yTicks)
    ax1.axes.set_xticks(range(1, 2**numQubits+1))
    ax1.axes.set_yticks(range(1, 2**numQubits+1))
    ax1.axes.set_zticks(np.arange(-1, 1.1, .2))
    ax1.axes.set_zlim3d(-1, 1)
    plt_given.title("Rho Real")
    fig.subplots_adjust(bottom = 0.2)
    ax1 = fig.add_axes([0.2, 0.10, 0.7, 0.065])
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap = colorMap,
                                    norm = norm,
                                    orientation = 'horizontal')

    # Imaginary graph
    fig = plt_given.figure()
    ax1 = fig.add_subplot(111, projection = '3d')
    dz = p.flatten().imag.astype(float)
    ax1.bar3d(xpos, ypos, zpos, dx, dy, dz, color = colorMap((dz + 1) / 2), edgecolor = "black", alpha = .8)

    ax1.axes.set_xticklabels(xTicks)
    ax1.axes.set_yticklabels(yTicks)
    ax1.axes.set_xticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_yticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_zticks(np.arange(-1, 1.1, .2))
    ax1.axes.set_zlim3d(-1, 1)
    plt_given.title("Rho Imaginary")

    fig.subplots_adjust(bottom = 0.2)
    ax1 = fig.add_axes([0.2, 0.10, 0.7, 0.065])
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap = colorMap,
                                    norm = norm,
                                    orientation = 'horizontal')

"""
saveRhoImages(p, pathToDirectory, customColor)
Desc: Creates and saves matlab plots of the density matrix.

Parameters
----------
p : ndarray with shape = (n, 2^numQubits, 2^numQubits)
    The density matrix you want to create plots of.
pathToDirectory : string
    Path to where you want your images to be saved.

See Also
 ------ 
makeRhoImages
"""
def saveRhoImages(p, pathToDirectory):
    # Set up
    numQubits = int(np.log2(p.shape[0]))
    xpos = np.zeros_like(p.flatten(), dtype = float)
    ypos = np.zeros_like(p.flatten(), dtype = float)
    for i in range(0, 2 ** numQubits):
        xpos[i * 2 ** numQubits:(1 + i) * 2 ** numQubits] = .5 + i
    for i in range(0, 2 ** numQubits):
        ypos[i::2 ** numQubits] = .5 + i
    zpos = np.zeros_like(p.flatten(), dtype = float)
    # width of cols
    dx = .9 * np.ones_like(xpos)
    dy = .9 * np.ones_like(ypos)
    # custom color map
    n_bin = 100
    cmap_name = 'my_list'
    colors = [(1 / 255.0, 221 / 255.0, 137 / 255.0),
              (32 / 255.0, 151 / 255.0, 138 / 255.0),
              (53 / 255.0, 106 / 255.0, 138 / 255.0),
              (86 / 255.0, 33 / 255.0, 139 / 255.0),
              (131 / 255.0, 75 / 255.0, 114 / 255.0),
              (173 / 255.0, 114 / 255.0, 90 / 255.0),
              (253 / 255.0, 187 / 255.0, 45 / 255.0)]
    colorMap = LinearSegmentedColormap.from_list(cmap_name, colors, N = n_bin)

    norm = mpl.colors.Normalize(vmin = -1, vmax = 1)

    tickBase = ["H", "V"]
    tick = [""]
    for x in range(numQubits):
        newTick = np.zeros(len(tick) * 2, dtype = "O")
        for i in range(len(tick)):
            for j in range(len(tickBase)):
                newTick[len(tick) * i + j] = tick[i] + tickBase[j]
        tick = newTick
    xTicks = ["|" + x + ">" for x in tick]
    yTicks = ["|" + x + ">" for x in tick]

    # Real Graph
    fig = plt.figure()
    ax1 = fig.add_subplot(111, projection = '3d')
    dz = p.flatten().astype(float)
    img = ax1.bar3d(xpos, ypos, zpos, dx, dy, dz, color = colorMap((dz + 1) / 2), edgecolor = "black", alpha = .8)

    ax1.axes.set_xticklabels(xTicks)
    ax1.axes.set_yticklabels(yTicks)
    ax1.axes.set_xticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_yticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_zticks(np.arange(-1, 1.1, .2))
    ax1.axes.set_zlim3d(-1, 1)
    plt.title("Rho Real")
    fig.subplots_adjust(bottom = 0.2)
    ax1 = fig.add_axes([0.2, 0.10, 0.7, 0.065])
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap = colorMap,
                                    norm = norm,
                                    orientation = 'horizontal')
    plt.savefig(pathToDirectory + "/rhobarReal.png", bbox_inches = 'tight', pad_inches = 0)

    # Imaginary graph
    fig = plt.figure()
    ax1 = fig.add_subplot(111, projection = '3d')
    dz = p.flatten().imag.astype(float)
    ax1.bar3d(xpos, ypos, zpos, dx, dy, dz, color = colorMap((dz + 1) / 2), edgecolor = "black", alpha = .8)

    ax1.axes.set_xticklabels(xTicks)
    ax1.axes.set_yticklabels(yTicks)
    ax1.axes.set_xticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_yticks(range(1, 2 ** numQubits + 1))
    ax1.axes.set_zticks(np.arange(-1, 1.1, .2))
    ax1.axes.set_zlim3d(-1, 1)
    plt.title("Rho Imaginary")
    fig.subplots_adjust(bottom = 0.2)
    ax1 = fig.add_axes([0.2, 0.10, 0.7, 0.065])
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap = colorMap,
                                    norm = norm,
                                    orientation = 'horizontal')

    plt.savefig(pathToDirectory + "/rhobarImag.png", bbox_inches = 'tight', pad_inches = 0)


"""
printLastOutput(tomo,bounds)
Desc: Prints the properties of the last tomography to the console. Properties are defined in tomography conf settings. The calculated properties are determined by self.err_functions.

Parameters
----------
tomo : Tomography
    The tomography object you want to see the output of.
bounds : int (optional)
    The number of monte carlo runs you want to perform to get a better estimate of each property. Default will use whatever is set in the conf settings.
"""
def printLastOutput(tomo, bounds = -1):
    p = np.array(tomo.last_rho.copy(), dtype = "O")
    print("State: ")
    mx = 0
    for i in range(p.shape[0]):
        for j in range(p.shape[1]):
            p[i, j] = floatToString(p[i, j]).replace(" ", "") + "  "
            if(len(p[i, j])>mx):
                mx = len(p[i, j])

    for i in range(p.shape[0]):
        for j in range(p.shape[1]):
            print(p[i, j] + " "*(mx-len(p[i, j])), end = "")
        print("")

    # print(p)
    properties = tomo.getProperties(bounds)
    for prop in properties:
        if(len(prop) >=3) and prop[2] != 'NA':
            print(prop[0] + " : " + floatToString(prop[1]) + " +/- " + floatToString(prop[2]))
        else:
            print(prop[0] + " : " + floatToString(prop[1]))

"""
matrixToHTML(M)
Desc: Creates an HTML table based on the given matrix.

Parameters
----------
M : 2d numpy array with shape = (2^numQubits, 2^numQubits)
    Matrix you would like to display on your html page.
printEigenVals : boolean
    Specify if you want eigen values to be calculated and displayed at the bottom of the table.
Returns
-------
res : string
    HTML code of the created table.

See Also
 ------ 
propertiesToHTML
"""
def matrixToHTML(M, printEigenVals = False):
    s = np.shape(M)
    res = '<table class="KwiatDataMatrix" style = \"border: 1px solid black;border-collapse: collapse;font-size: 15px; table-layout:fixed;width:100%;margin-top: 25px;\">'
    for i in range(s[0]):
        res = res+' <tr>'
        for j in range(s[1]):
            # res = res + '<td style = "border: 1px solid black;">' + str(np.real(M[i, j])) + "<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">+</div><BR>"+ str(np.imag(M[i, j]))
            # res = res + '<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">j</div></td>'
            res = res + '<td style = "border: 1px solid black;">' + floatToString(M[i, j], True)+ '</td>'
        res = res +'</tr>'
    res = res+'</table>'
    if(printEigenVals):
        d, v = np.linalg.eig(M)
        sum = 0
        eigenVals = "<h5>Eigenvalues: "
        for x in range(0, len(d)):
            eigenVals = eigenVals+str(round(d[x].real, 5))
            if(abs(d[x].imag)>.00001):
                eigenVals = eigenVals+"< div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">+</div>"
                eigenVals = eigenVals+str(round(d[x].imag, 5))
                eigenVals = eigenVals+"<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">j</div>"
            eigenVals = eigenVals+" , "
            sum+= d[x]
        eigenVals = str(eigenVals)[0:len(str(eigenVals))-2]
        eigenVals = eigenVals +"</h5>"
        res = res+eigenVals
    return res


"""
propertiesToHTML(vals)
Desc: Creates an HTML table based on the given property values.

Parameters
----------
vals : ndarray with shape = (length of self.err_functions, 2)
    The first col is the name of the property.
    The second col is the value of the property.
    The third col is the error bound on the property.

Returns
-------
res : string
    HTML code of the created table.

See Also
 ------ 
matrixToHTML
"""
def propertiesToHTML(vals):
    f = '<h3 >Properties of Rho</h3><table style = \"width:60%;margin-top:10px;font-size: 15px;padding-bottom:5px;float:none;\"><tr><td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;">Property</td>'
    hasSTD = len(vals) > 2
    if hasSTD and any(vals[:,2] != "NA"):
        f += '<td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;">Average Value</td>'
        f += '<td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;">STD Error</td></tr>'
        f += '<td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;"></td></tr>'
    else:
        f += '<td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;">Value</td>'
        f += '<td style = "font-size: 20;font-weight: 1000;color: rebeccapurple;padding-bottom:5px;"></td></tr>'
    for v in vals:
        if(v[1] != "NA"):
            f += '<tr>'
            f += '<td><div onmouseover = "Tip( ' + v[0].replace(" ", "") + 'Tip)" onmouseout = "hideTip()"><a style="cursor:help;">'+v[0] + '</a></td>'
            f += '<td name = "' + v[0].replace(" ", "") + '_value">' + floatToString(v[1], True) + '</td>'
            if hasSTD and v[2] != "NA":
                f += '<td> +/- ' + floatToString(v[2], True) + '</td>'
            else:
                f += "<td></td>"
            f += '</tr>'
    f += "</table>"
    return f


"""
stateToString(vals)
Desc: Creates a string of the 1d pure state.

Parameters
----------    
pure_state : 1darray with length = 2
    The state in ket form.
"""
def stateToString(state):
    state = removeGlobalPhase(state)
    return floatToString(state[0])+"|H> + "+floatToString(state[1])+"|V>"



from __future__ import print_function
import numpy as np
import numpy.random as rand
"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""



# These are various helper functions used in other main functions in the TomoFunction file


# Helper function for partial Transpose in TomoFunctions
def partial_transpose_helper(m, d):
    if m.shape[0] == d:
        val = m.transpose()
    else:
        na = int(d)
        nb = int(len(m)/d)
        y = np.zeros([nb, nb, na, na])+0j
        val = np.zeros([len(m), len(m)])+0j
        for j in range(nb):
            for k in range(nb):
                y[j, k] = m[j*nb:j*nb+na, k*nb:k*nb+na]
        for j in range(nb):
            for k in range(nb):
                val[j*nb:j*nb+na, k*nb:k*nb+na] = y[k, j]

    return val

"""
    random_ginibre(D)
    Desc: Returns a random matrix from the Ginibre ensemble of size DxD. 
    This is a complex matrix whos elements are a+ib | a,b iid. Norm(0,1)

    Parameters
    ----------
    D :int
        The dimension of the Matrix

    Returns
    -------
    mat : ndarray with shape = (2^N, 2^N)
        The random matrix
    """
def random_ginibre(D=2):
    mat = np.zeros((D, D), dtype=complex)
    for i in range(D):
        for j in range(D):
            mat[i, j] = rand.normal(0, 1) + rand.normal(0, 1) * 1j

    return mat


def phaserToComplex(phaser):
    Magn = phaser[0]
    Phase = phaser[1]
    complex = Magn*(np.cos(Phase) +  1j*np.sin(Phase))
    return complex

def complexToPhaser(complex):
    Magn = np.absolute(complex)
    Phase = np.angle(complex)
    if(Phase<0):
        Phase += 2*np.pi
    return np.array([Magn,Phase])

def isStateVector(state):
    return np.ndim(state) == 1


from __future__ import print_function
import numpy as np

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""


# These are various helper functions used in other main functions in the TomoDisplay file


# This function rounds and converts floats to strings. It handles complex numbers and uses scientific notation
def floatToString(x, html = False):
    if(x == "NA"):
        return x
    if(abs(x.imag) > 10**-8):
        tempIMAG = floatToString(x.imag)
        if(abs(x.real) > 10**-8):
            if(html):
                if(tempIMAG[0] != "-"):
                    return floatToString(x.real) + "<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\"> + i</div>" + tempIMAG
                else:
                    return floatToString(x.real) + "<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\"> - i</div>" + tempIMAG[1:]
            else:
                if (tempIMAG[0] != "-"):
                    return floatToString(x.real) + " + i" + tempIMAG
                else:
                    return floatToString(x.real) + " - i" + tempIMAG[1:]
        else:
            if (html):
                if (tempIMAG[0] != "-"):
                    return "<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">i</div>" + tempIMAG
                else:
                    return "<div style = \"color:rebeccapurple;font-weight: bold;display:inline;\">-i</div>" + tempIMAG[1:]
            else:
                if (tempIMAG[0] != "-"):
                    return "i" + tempIMAG
                else:
                    return "-i" + tempIMAG[1:]
    else:
        if(x == float("inf")):
            return "inf"
        if isNaN(x):
            return "nan"
        if (abs(x.real) <= 10 ** -8):
            return "0"

        s = "{:e}".format(float(x.real))
        [num, power] = s.split("e")
        if(num[0] == "-"):
            num = num[:5]
        else:
            num = num[:4]

        if(abs(float(power)) > 2):
            return num+"e"+power
        else:
            s = float(num)*10**float(power)
            s = np.around(s, 2-int(power))
            s = str(s)
            return s
# checks if a number is nan
def isNaN(num):
    return num != num
#
# tests = [1*10**-20, 1j*10**-20, 1*10**-20+1j*10**-20, 1*10**-5+1j*10**-20, 1*10**-20+1j*10**-5, 1*10**-5, 1j*10**-5, 0]
# for i in tests:
#     print(str(i)+" : "+ floatToString(i))



from setuptools import setup

"""
Copyright 2020 University of Illinois Board of Trustees.
Licensed under the terms of an MIT license
"""


"""CHECK OUT THE REFERENCE PAGE ON OUR WEBSITE :
https://quantumtomo.web.illinois.edu/Doc/"""

def readme():
    with open('README.md') as f:
        README  = f.read()
    return README


setup(
    name = "Quantum-Tomography",
    version = "1.0.5.0",
    description = "A python library to help perform tomography on a quantum state.",
    long_description = readme(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/KwiatQIM/Quantum-Tomography",
    author = "University of Illinois",
    author_email = "turroscott@gmail.com",
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    package_dir = {'': 'src'},
    packages = ["QuantumTomography"],
    include_package_data = True,
    install_requires = ["scipy", "numpy", "matplotlib"],
    entry_points = {
        "console_scripts": [
            "Quantum-Tomography = QuantumTomography.cli:main",
        ]
    },
    project_urls = {
      'Documentation': 'https://quantumtomo.web.illinois.edu/Doc/',
      'Video Tutorial': 'https://www.youtube.com/watch?v=I-214P0LOfQ&list=PLJLHMKtk5Pqy9w9aCuyowUF1p7pl2JCI9&index=3'
    }
)
